import numpy as np
import simsoptpp as sopp


class Area(object):
    """
    Wrapper class for surface area label.
    """

    def __init__(self, surface):
        self.surface = surface

    def J(self):
        """
        Compute the area of a surface.
        """
        return self.surface.area()

    def dJ_by_dsurfacecoefficients(self):
        """
        Calculate the derivatives with respect to the surface coefficients.
        """
        return self.surface.darea_by_dcoeff()

    def d2J_by_dsurfacecoefficientsdsurfacecoefficients(self):
        """
        Calculate the second derivatives with respect to the surface coefficients.
        """
        return self.surface.d2area_by_dcoeffdcoeff()


class Volume(object):
    """
    Wrapper class for volume label.
    """

    def __init__(self, surface):
        self.surface = surface

    def J(self):
        """
        Compute the volume enclosed by the surface.
        """
        return self.surface.volume()

    def dJ_by_dsurfacecoefficients(self):
        """
        Calculate the derivatives with respect to the surface coefficients.
        """
        return self.surface.dvolume_by_dcoeff()

    def d2J_by_dsurfacecoefficientsdsurfacecoefficients(self):
        """
        Calculate the second derivatives with respect to the surface coefficients.
        """
        return self.surface.d2volume_by_dcoeffdcoeff()


class ToroidalFlux(object):
    r"""
    Given a surface and Biot Savart kernel, this objective calculates

    .. math::
       J &= \int_{S_{\varphi}} \mathbf{B} \cdot \mathbf{n} ~ds, \\
       &= \int_{S_{\varphi}} \text{curl} \mathbf{A} \cdot \mathbf{n} ~ds, \\
       &= \int_{\partial S_{\varphi}} \mathbf{A} \cdot \mathbf{t}~dl,

    where :math:`S_{\varphi}` is a surface of constant :math:`\varphi`, and :math:`\mathbf A`
    is the magnetic vector potential.
    """

    def __init__(self, surface, biotsavart, idx=0):
        self.surface = surface
        self.biotsavart = biotsavart
        self.idx = idx
        self.surface.dependencies.append(self)
        self.invalidate_cache()

    def invalidate_cache(self):
        x = self.surface.gamma()[self.idx]
        self.biotsavart.set_points(x)

    def J(self):
        r"""
        Compute the toroidal flux on the surface where
        :math:`\varphi = \texttt{quadpoints_varphi}[\texttt{idx}]`.
        """
        xtheta = self.surface.gammadash2()[self.idx]
        ntheta = self.surface.gamma().shape[1]
        A = self.biotsavart.A()
        tf = np.sum(A * xtheta)/ntheta
        return tf

    def dJ_by_dsurfacecoefficients(self):
        """
        Calculate the derivatives with respect to the surface coefficients.
        """
        ntheta = self.surface.gamma().shape[1]
        dA_by_dX = self.biotsavart.dA_by_dX()
        A = self.biotsavart.A()
        dgammadash2 = self.surface.gammadash2()[self.idx, :]
        dgammadash2_by_dc = self.surface.dgammadash2_by_dcoeff()[self.idx, :]

        dx_dc = self.surface.dgamma_by_dcoeff()[self.idx]
        dA_dc = np.sum(dA_by_dX[..., :, None] * dx_dc[..., None, :], axis=1)
        term1 = np.sum(dA_dc * dgammadash2[..., None], axis=(0, 1))
        term2 = np.sum(A[..., None] * dgammadash2_by_dc, axis=(0, 1))

        out = (term1+term2)/ntheta
        return out

    def d2J_by_dsurfacecoefficientsdsurfacecoefficients(self):
        """
        Calculate the second derivatives with respect to the surface coefficients.
        """
        ntheta = self.surface.gamma().shape[1]
        dx_dc = self.surface.dgamma_by_dcoeff()[self.idx]
        d2A_by_dXdX = self.biotsavart.d2A_by_dXdX().reshape((ntheta, 3, 3, 3))
        dA_by_dX = self.biotsavart.dA_by_dX()
        dA_dc = np.sum(dA_by_dX[..., :, None] * dx_dc[..., None, :], axis=1)
        d2A_dcdc = np.einsum('jkpl,jpn,jkm->jlmn', d2A_by_dXdX, dx_dc, dx_dc)

        dgammadash2 = self.surface.gammadash2()[self.idx]
        dgammadash2_by_dc = self.surface.dgammadash2_by_dcoeff()[self.idx]

        term1 = np.sum(d2A_dcdc * dgammadash2[..., None, None], axis=-3)
        term2 = np.sum(dA_dc[..., :, None] * dgammadash2_by_dc[..., None, :], axis=-3)
        term3 = np.sum(dA_dc[..., None, :] * dgammadash2_by_dc[..., :, None], axis=-3)

        out = (1/ntheta) * np.sum(term1+term2+term3, axis=0)
        return out


def boozer_surface_residual(surface, iota, G, biotsavart, derivatives=0):
    r"""
    For a given surface, this function computes the
    residual

    .. math::
        G\mathbf B_\text{BS}(\mathbf x) - ||\mathbf B_\text{BS}(\mathbf x)||^2  (\mathbf x_\varphi + \iota  \mathbf x_\theta)

    as well as the derivatives of this residual with respect to surface dofs,
    iota, and G.  In the above, :math:`\mathbf x` are points on the surface, :math:`\iota` is the
    rotational transform on that surface, and :math:`\mathbf B_{\text{BS}}` is the magnetic field
    computed using the Biot-Savart law.

    :math:`G` is known for exact boozer surfaces, so if ``G=None`` is passed, then that
    value is used instead.
    """

    user_provided_G = G is not None
    if not user_provided_G:
        G = 2. * np.pi * np.sum(np.abs(biotsavart.coil_currents)) * (4 * np.pi * 10**(-7) / (2 * np.pi))

    x = surface.gamma()
    xphi = surface.gammadash1()
    xtheta = surface.gammadash2()
    nphi = x.shape[0]
    ntheta = x.shape[1]

    xsemiflat = x.reshape((x.size//3, 3)).copy()

    biotsavart.set_points(xsemiflat)

    biotsavart.compute(derivatives)
    B = biotsavart.B().reshape((nphi, ntheta, 3))

    tang = xphi + iota * xtheta
    B2 = np.sum(B**2, axis=2)
    residual = G*B - B2[..., None] * tang

    residual_flattened = residual.reshape((nphi*ntheta*3, ))
    r = residual_flattened
    if derivatives == 0:
        return r,

    dx_dc = surface.dgamma_by_dcoeff()
    dxphi_dc = surface.dgammadash1_by_dcoeff()
    dxtheta_dc = surface.dgammadash2_by_dcoeff()
    nsurfdofs = dx_dc.shape[-1]

    dB_by_dX = biotsavart.dB_by_dX().reshape((nphi, ntheta, 3, 3))
    dB_dc = np.einsum('ijkl,ijkm->ijlm', dB_by_dX, dx_dc)

    # dresidual_dc = G*dB_dc - 2*np.sum(B[..., None]*dB_dc, axis=2)[:, :, None, :] * tang[..., None] - B2[..., None, None] * (dxphi_dc + iota * dxtheta_dc)
    dresidual_dc = sopp.boozer_dresidual_dc(G, dB_dc, B, tang, B2, dxphi_dc, iota, dxtheta_dc)
    dresidual_diota = -B2[..., None] * xtheta

    dresidual_dc_flattened = dresidual_dc.reshape((nphi*ntheta*3, nsurfdofs))
    dresidual_diota_flattened = dresidual_diota.reshape((nphi*ntheta*3, 1))

    if user_provided_G:
        dresidual_dG = B
        dresidual_dG_flattened = dresidual_dG.reshape((nphi*ntheta*3, 1))
        J = np.concatenate((dresidual_dc_flattened, dresidual_diota_flattened, dresidual_dG_flattened), axis=1)
    else:
        J = np.concatenate((dresidual_dc_flattened, dresidual_diota_flattened), axis=1)
    if derivatives == 1:
        return r, J

    d2B_by_dXdX = biotsavart.d2B_by_dXdX().reshape((nphi, ntheta, 3, 3, 3))
    d2B_dcdc = np.einsum('ijkpl,ijpn,ijkm->ijlmn', d2B_by_dXdX, dx_dc, dx_dc)
    dB2_dc = 2. * np.einsum('ijl,ijlm->ijm', B, dB_dc)

    term1 = np.einsum('ijlm,ijln->ijmn', dB_dc, dB_dc)
    term2 = np.einsum('ijlmn,ijl->ijmn', d2B_dcdc, B)
    d2B2_dcdc = 2*(term1 + term2)

    term1 = -(dxphi_dc[..., None, :] + iota * dxtheta_dc[..., None, :]) * dB2_dc[..., None, :, None]
    term2 = -(dxphi_dc[..., :, None] + iota * dxtheta_dc[..., :, None]) * dB2_dc[..., None, None, :]
    term3 = -(xphi[..., None, None] + iota * xtheta[..., None, None]) * d2B2_dcdc[..., None, :, :]
    d2residual_by_dcdc = G * d2B_dcdc + term1 + term2 + term3
    d2residual_by_dcdiota = -(dB2_dc[..., None, :] * xtheta[..., :, None] + B2[..., None, None] * dxtheta_dc)
    d2residual_by_diotadiota = np.zeros(dresidual_diota.shape)

    d2residual_by_dcdc_flattened = d2residual_by_dcdc.reshape((nphi*ntheta*3, nsurfdofs, nsurfdofs))
    d2residual_by_dcdiota_flattened = d2residual_by_dcdiota.reshape((nphi*ntheta*3, nsurfdofs))
    d2residual_by_diotadiota_flattened = d2residual_by_diotadiota.reshape((nphi*ntheta*3,))

    if user_provided_G:
        d2residual_by_dcdG = dB_dc
        d2residual_by_diotadG = np.zeros(dresidual_diota.shape)
        d2residual_by_dGdG = np.zeros(dresidual_dG.shape)
        d2residual_by_dcdG_flattened = d2residual_by_dcdG.reshape((nphi*ntheta*3, nsurfdofs))
        d2residual_by_diotadG_flattened = d2residual_by_diotadG.reshape((nphi*ntheta*3,))
        d2residual_by_dGdG_flattened = d2residual_by_dGdG.reshape((nphi*ntheta*3,))
        H = np.zeros((nphi*ntheta*3, nsurfdofs + 2, nsurfdofs + 2))
        # noqa turns out linting so that we can align everything neatly
        H[:, :nsurfdofs, :nsurfdofs] = d2residual_by_dcdc_flattened        # noqa (0, 0) dcdc
        H[:, :nsurfdofs, nsurfdofs] = d2residual_by_dcdiota_flattened     # noqa (0, 1) dcdiota
        H[:, :nsurfdofs, nsurfdofs+1] = d2residual_by_dcdG_flattened        # noqa (0, 2) dcdG
        H[:, nsurfdofs, :nsurfdofs] = d2residual_by_dcdiota_flattened     # noqa (1, 0) diotadc
        H[:, nsurfdofs, nsurfdofs] = d2residual_by_diotadiota_flattened  # noqa (1, 1) diotadiota
        H[:, nsurfdofs, nsurfdofs+1] = d2residual_by_diotadiota_flattened  # noqa (1, 2) diotadG
        H[:, nsurfdofs+1, :nsurfdofs] = d2residual_by_dcdG_flattened        # noqa (2, 0) dGdc
        H[:, nsurfdofs+1, nsurfdofs] = d2residual_by_diotadG_flattened     # noqa (2, 1) dGdiota
        H[:, nsurfdofs+1, nsurfdofs+1] = d2residual_by_dGdG_flattened        # noqa (2, 2) dGdG
    else:
        H = np.zeros((nphi*ntheta*3, nsurfdofs + 1, nsurfdofs + 1))

        H[:, :nsurfdofs, :nsurfdofs] = d2residual_by_dcdc_flattened        # noqa (0, 0) dcdc
        H[:, :nsurfdofs, nsurfdofs] = d2residual_by_dcdiota_flattened     # noqa (0, 1) dcdiota
        H[:, nsurfdofs, :nsurfdofs] = d2residual_by_dcdiota_flattened     # noqa (1, 0) diotadc
        H[:, nsurfdofs, nsurfdofs] = d2residual_by_diotadiota_flattened  # noqa (1, 1) diotadiota

    return r, J, H


class QfmResidual(object):
    r"""
    For a given surface :math:`S`, this class computes the residual

    .. math::
        f(S) = \frac{\int_{S} d^2 x \, (\textbf{B} \cdot \hat{\textbf{n}})^2}{\int_{S} d^2 x \, B^2}

    where :math:`\textbf{B}` is the magnetic field from :mod:`biotsavart`,
    :math:`\hat{\textbf{n}}` is the unit normal on a given surface, and the
    integration is performed over the surface. Derivatives are computed wrt the
    surface dofs.
    """

    def __init__(self, surface, biotsavart):
        self.surface = surface
        self.biotsavart = biotsavart
        self.surface.dependencies.append(self)
        self.invalidate_cache()

    def invalidate_cache(self):
        x = self.surface.gamma()
        xsemiflat = x.reshape((x.size//3, 3)).copy()
        self.biotsavart.set_points(xsemiflat)

    def J(self):
        N = self.surface.normal()
        norm_N = np.linalg.norm(N, axis=2)
        n = N/norm_N[:, :, None]
        x = self.surface.gamma()
        nphi = x.shape[0]
        ntheta = x.shape[1]
        B = self.biotsavart.B().reshape((nphi, ntheta, 3))
        B_n = np.sum(B * n, axis=2)
        norm_B = np.linalg.norm(B, axis=2)
        return np.sum(B_n**2 * norm_N)/np.sum(norm_B**2 * norm_N)

    def dJ_by_dsurfacecoefficients(self):
        """
        Calculate the derivatives with respect to the surface coefficients
        """
        x = self.surface.gamma()
        nphi = x.shape[0]
        ntheta = x.shape[1]
        dx_by_dc = self.surface.dgamma_by_dcoeff()
        dB_by_dX = self.biotsavart.dB_by_dX().reshape((nphi, ntheta, 3, 3))
        B = self.biotsavart.B().reshape((nphi, ntheta, 3))
        d_B = np.einsum('ijkl,ijkm->ijlm', dB_by_dX, dx_by_dc)

        N = self.surface.normal()
        norm_N = np.linalg.norm(N, axis=2)
        n = N/norm_N[:, :, None]
        d_N = self.surface.dnormal_by_dcoeff()
        d_norm_N = np.einsum('ijkl,ijk->ijl', d_N, n)
        d_n = d_N/norm_N[:, :, None, None] \
            - N[:, :, :, None] * d_norm_N[:, :, None, :]/norm_N[:, :, None, None]**2

        B_n = np.sum(B * n, axis=2)
        norm_B = np.linalg.norm(B, axis=2)

        d_B_n = np.einsum('ijkl,ijk->ijl', d_B, n) + np.einsum('ijk,ijkl->ijl', B, d_n)
        d_norm_B = np.einsum('ijkl,ijk->ijl', d_B, B/norm_B[:, :, None])

        num = np.sum(B_n**2 * norm_N)
        denom = np.sum(norm_B**2 * norm_N)
        d_num = np.sum(2 * d_B_n * B_n[:, :, None] * norm_N[:, :, None]
                       + B_n[:, :, None]**2 * d_norm_N, axis=(0, 1))
        d_denom = np.sum(2 * d_norm_B * norm_B[:, :, None] * norm_N[:, :, None]
                         + norm_B[:, :, None]**2 * d_norm_N, axis=(0, 1))
        return d_num/denom - d_denom*num/(denom*denom)

    def d2J_by_dsurfacecoefficientsdsurfacecoefficients(self):
        """
        Calculate the second derivative wrt the surface coefficients
        """
        x = self.surface.gamma()
        nphi = x.shape[0]
        ntheta = x.shape[1]
        dB_by_dX = self.biotsavart.dB_by_dX().reshape((nphi, ntheta, 3, 3))
        d2B_by_dXdX = self.biotsavart.d2B_by_dXdX().reshape((nphi, ntheta, 3, 3, 3))
        B = self.biotsavart.B().reshape((nphi, ntheta, 3))
        dx_by_dc = self.surface.dgamma_by_dcoeff()

        d_B = np.einsum('ijkl,ijkm->ijml', dx_by_dc, dB_by_dX)
        d2_B = np.einsum('ijkpl,ijpn,ijkm->ijlmn', d2B_by_dXdX, dx_by_dc, dx_by_dc)

        N = self.surface.normal()
        norm_N = np.linalg.norm(N, axis=2)
        n = N/norm_N[:, :, None]

        d_N = self.surface.dnormal_by_dcoeff()
        d2_N = self.surface.d2normal_by_dcoeffdcoeff()

        d_norm_N = np.einsum('ijkl,ijk->ijl', d_N, n)
        d2_norm_N = (np.einsum('imjk,imjl->imkl', d_N, d_N)
                     + np.einsum('imjkl,imj->imkl', d2_N, N)
                     - d_norm_N[:, :, :, None]*d_norm_N[:, :, None, :])/norm_N[:, :, None, None]
        d_n = d_N/norm_N[:, :, None, None] \
            - N[:, :, :, None] * d_norm_N[:, :, None, :]/norm_N[:, :, None, None]**2
        d2_n = d2_N/norm_N[:, :, None, None, None] \
            - np.einsum('imjk,iml->imjkl', d_N, d_norm_N)/norm_N[:, :, None, None, None]**2 \
            - np.einsum('imjk,iml->imjlk', d_N, d_norm_N)/norm_N[:, :, None, None, None]**2 \
            + 2 * np.einsum('imj,iml,imk->imjlk', N, d_norm_N, d_norm_N)/norm_N[:, :, None, None, None]**3 \
            - np.einsum('imj,imkl->imjkl', N, d2_norm_N)/norm_N[:, :, None, None, None]**2

        B_n = np.sum(B * n, axis=2)
        norm_B = np.linalg.norm(B, axis=2)

        d_B_n = np.einsum('ijkl,ijk->ijl', d_B, n) + np.einsum('ijk,ijkl->ijl', B, d_n)
        d_norm_B = np.einsum('ijkl,ijk->ijl', d_B, B/norm_B[:, :, None])

        d2_B_n = np.einsum('imjkl,imj->imkl', d2_B, n) \
            + np.einsum('imj,imjkl->imkl', B, d2_n) \
            + np.einsum('imjk,imjl->imkl', d_B, d_n) \
            + np.einsum('imjk,imjl->imlk', d_B, d_n)
        d2_norm_B = (np.einsum('imjkl,imj->imkl', d2_B, B)
                     + np.einsum('imjk,imjl->imkl', d_B, d_B)
                     - np.einsum('imk,iml->imkl', d_norm_B, d_norm_B))/norm_B[:, :, None, None]

        num = np.sum(B_n**2 * norm_N)
        denom = np.sum(norm_B**2 * norm_N)
        d_num = np.sum(2 * d_B_n * B_n[:, :, None] * norm_N[:, :, None]
                       + B_n[:, :, None]**2 * d_norm_N, axis=(0, 1))
        d2_num = np.sum(2*(d2_B_n * B_n[:, :, None, None] \
                           + d_B_n[:, :, :, None] * d_B_n[:, :, None, :]) * norm_N[:, :, None, None]
                        + 2 * d_B_n[:, :, None, :] * B_n[:, :, None, None] * d_norm_N[:, :, :, None]
                        + 2 * d_B_n[:, :, :, None] * B_n[:, :, None, None] * d_norm_N[:, :, None, :]
                        + B_n[:, :, None, None]**2 * d2_norm_N, axis=(0, 1))
        d_denom = np.sum(2 * d_norm_B * norm_B[:, :, None] * norm_N[:, :, None]
                         + norm_B[:, :, None]**2 * d_norm_N, axis=(0, 1))
        d2_denom = np.sum(2*(d2_norm_B * norm_B[:, :, None, None] \
                             + d_norm_B[:, :, :, None] * d_norm_B[:, :, None, :]) * norm_N[:, :, None, None]
                          + 2 * d_norm_B[:, :, None, :] * norm_B[:, :, None, None] * d_norm_N[:, :, :, None]
                          + 2 * d_norm_B[:, :, :, None] * norm_B[:, :, None, None] * d_norm_N[:, :, None, :]
                          + norm_B[:, :, None, None]**2 * d2_norm_N, axis=(0, 1))

        return d2_num/denom - d_num[:, None]*d_denom[None, :]/denom**2 \
            - d_num[None, :]*d_denom[:, None]/denom**2 \
            - d2_denom*num/(denom*denom) \
            + 2 * d_denom[:, None]*d_denom[None, :]*num/(denom**3)



from benchmark_xtensor_python import sum_array
import numpy as np

u = np.ones(1000000, dtype=float)
from timeit import timeit
print (timeit ('sum_array(u)', setup='from __main__ import u, sum_array', number=1000))



import numpy as np
import simsoptpp as sopp
from .surface import Surface
from .surfacerzfourier import SurfaceRZFourier


class SurfaceXYZFourier(sopp.SurfaceXYZFourier, Surface):
    r"""`SurfaceXYZFourier` is a surface that is represented in Cartesian
    coordinates using the following Fourier series:

    .. math::
        \hat x(\phi,\theta) &= \sum_{m=0}^{m_\text{pol}} \sum_{n=-n_{\text{tor}}}^{n_{tor}} [
              x_{c,m,n} \cos(m \theta - n_\text{ fp } n \phi)
            + x_{s,m,n} \sin(m \theta - n_\text{ fp } n \phi)]\\
        \hat y(\phi,\theta) &= \sum_{m=0}^{m_\text{pol}} \sum_{n=-n_\text{tor}}^{n_\text{tor}} [
              y_{c,m,n} \cos(m \theta - n_\text{fp} n \phi)
            + y_{s,m,n} \sin(m \theta - n_\text{fp} n \phi)]\\
        z(\phi,\theta) &= \sum_{m=0}^{m_\text{pol}} \sum_{n=-n_\text{tor}}^{n_\text{tor}} [
              z_{c,m,n} \cos(m \theta - n_\text{fp}n \phi)
            + z_{s,m,n} \sin(m \theta - n_\text{fp}n \phi)]

    where

    .. math::
        x &= \hat x \cos(\phi) - \hat y \sin(\phi)\\
        y &= \hat x \sin(\phi) + \hat y \cos(\phi)

    Note that for :math:`m=0` we skip the :math:`n<0` term for the cos terms, and the :math:`n \leq 0`
    for the sin terms.

    When enforcing stellarator symmetry, we set the

    .. math::
        x_{s,*,*}, ~y_{c,*,*}, \text{and} ~z_{c,*,*}

    terms to zero.
    """

    def __init__(self, nfp=1, stellsym=True, mpol=1, ntor=0, quadpoints_phi=32, quadpoints_theta=32):
        if isinstance(quadpoints_phi, np.ndarray):
            quadpoints_phi = list(quadpoints_phi)
            quadpoints_theta = list(quadpoints_theta)
        sopp.SurfaceXYZFourier.__init__(self, mpol, ntor, nfp, stellsym, quadpoints_phi, quadpoints_theta)
        self.xc[0, ntor] = 1.0
        self.xc[1, ntor] = 0.1
        self.zs[1, ntor] = 0.1
        Surface.__init__(self)

    def get_dofs(self):
        """
        Return the dofs associated to this surface.
        """
        return np.asarray(sopp.SurfaceXYZFourier.get_dofs(self))

    def to_RZFourier(self):
        """
        Return a SurfaceRZFourier instance corresponding to the shape of this
        surface.
        """
        ntor = self.ntor
        mpol = self.mpol 
        surf = SurfaceRZFourier(nfp=self.nfp, 
                                stellsym=self.stellsym, 
                                mpol=mpol, 
                                ntor=ntor, 
                                quadpoints_phi=self.quadpoints_phi, 
                                quadpoints_theta=self.quadpoints_theta)

        gamma = np.zeros((surf.quadpoints_phi.size, surf.quadpoints_theta.size, 3))
        for idx in range(gamma.shape[0]):
            gamma[idx, :, :] = self.cross_section(surf.quadpoints_phi[idx]*2*np.pi)

        surf.least_squares_fit(gamma)
        return surf

    def set_dofs(self, dofs):
        """
        Set the dofs associated to this surface.
        """
        sopp.SurfaceXYZFourier.set_dofs(self, dofs)
        for d in self.dependencies:
            d.invalidate_cache()



# coding: utf-8
# Copyright (c) HiddenSymmetries Development Team.
# Distributed under the terms of the LGPL License

"""
This module provides two main functions, fd_jac_mpi and
least_squares_mpi_solve. Also included are some functions that help in
the operation of these main functions.
"""

import logging
from datetime import datetime
from time import time

import numpy as np
from scipy.optimize import least_squares

try:
    from mpi4py import MPI
except ImportError as err:
    MPI = None

from .._core.graph_optimizable import Optimizable
from ..util.mpi import MpiPartition
from ..util.types import RealArray
from ..objectives.graph_least_squares import LeastSquaresProblem

logger = logging.getLogger(__name__)

# Constants for signaling to workers what task to do:
CALCULATE_F = 1
CALCULATE_JAC = 2
CALCULATE_FD_JAC = 3


def _mpi_leaders_task(mpi: MpiPartition,
                      prob: Optimizable,
                      data: int):
    """
    This function is called by group leaders when
    MpiPartition.leaders_loop() receives a signal to do something.

    We have to take a "data" argument, but there is only 1 task we
    would do, so we don't use it.

    Args:
        mpi: A :obj:`simsopt.util.mpi.MpiPartition` object, storing
          the information about how the pool of MPI processes is
          divided into worker groups.
        prob: Optimizable object
        data: Dummy argument for this function
    """
    logger.debug('mpi leaders task')

    # x is a buffer for receiving the state vector:
    x = np.empty(prob.dof_size, dtype='d')
    # If we make it here, we must be doing a fd_jac_par
    # calculation, so receive the state vector: mpi4py has
    # separate bcast and Bcast functions!!  comm.Bcast(x,
    # root=0)
    x = mpi.comm_leaders.bcast(x, root=0)
    logger.debug(f'mpi leaders loop x={x}')
    prob.x = x
    fd_jac_mpi(prob, mpi)


def _mpi_workers_task(mpi: MpiPartition,
                      prob: Optimizable,
                      data: str):
    """
    This function is called by worker processes when
    MpiPartition.workers_loop() receives a signal to do something.

    Args:
        mpi: A :obj:`simsopt.util.mpi.MpiPartition` object, storing
          the information about how the pool of MPI processes is
          divided into worker groups.
        prob: Optimizable object
        data: Integer with a value from 1 to 3
    """
    logger.debug('mpi workers task')

    # x is a buffer for receiving the state vector:
    x = np.empty(prob.dof_size, dtype='d')
    # If we make it here, we must be doing a fd_jac_par
    # calculation, so receive the state vector: mpi4py has
    # separate bcast and Bcast functions!!  comm.Bcast(x, root=0)
    x = mpi.comm_groups.bcast(x, root=0)
    logger.debug('worker loop worker x={}'.format(x))
    prob.x = x

    # We don't store or do anything with f() or jac(), because
    # the group leader will handle that.
    if data == CALCULATE_F:
        try:
            prob.objective()
        except:
            logger.info("Exception caught by worker during dofs.f()")

    elif data == CALCULATE_JAC:
        try:
            prob.jac()
        except:
            logger.info("Exception caught by worker during dofs.jac()")

    else:
        raise ValueError('Unexpected data in worker_loop')


def fd_jac_mpi(prob: Optimizable,
               mpi: MpiPartition,
               x: RealArray = None,
               eps: float = 1e-7,
               centered: bool = False):
    """
    Compute the finite-difference Jacobian of the functions in dofs
    with respect to all non-fixed degrees of freedom. Parallel
    function evaluations will be used.

    If the argument x is not supplied, the Jacobian will be
    evaluated for the present state vector. If x is supplied, then
    first get_dofs() will be called for each object to set the
    global state vector to x.

    There are 2 ways to call this function. In method 1, all procs
    (including workers) call this function (so mpi.is_apart is
    False). In this case, the worker loop will be started
    automatically. In method 2, the worker loop has already been
    started before this function is called, as would be the case
    in least_squares_mpi_solve(). Then only the group leaders
    call this function.

    Args:
        prob: The map from :math:`\mathbb{R}^n \\to \mathbb{R}^m` for which you
          want to compute the Jacobian.
        mpi: A :obj:`simsopt.util.mpi.MpiPartition` object, storing
          the information about how the pool of MPI processes is
          divided into worker groups.
        x: The 1D state vector at which you wish to evaluate the Jacobian.
          If ``None``, the Jacobian will be evaluated at the present
          state vector.
        eps: Step size for finite differences.
        centered: If ``True``, centered finite differences will be used.
          If ``false``, one-sided finite differences will be used.

    Returns:
        tuple containing

        - **jac** (*numpy.ndarray*) -- The Jacobian matrix.
        - **xmat** (*numpy.ndarray*) -- A matrix, the columns of which give
          all the values of x at which the functions were evaluated.
        - **fmat** (*numpy.ndarray*) -- A matrix, the columns of which give
          the corresponding values of the functions.
    """
    if MPI is None:
        raise RuntimeError("fd_jac_mpi requires the mpi4py package.")

    apart_at_start = mpi.is_apart
    if not apart_at_start:
        mpi.worker_loop(lambda mpi2, data: _mpi_workers_task(mpi2, prob, data))
    if not mpi.proc0_groups:
        return (None, None, None)

    # Only group leaders execute this next section.

    if x is not None:
        prob.x = x

    logger.info(f'Beginning parallel finite difference gradient calculation')

    x0 = prob.x
    # Make sure all leaders have the same x0.
    mpi.comm_leaders.Bcast(x0)
    #logger.info('  nparams: {}, nfuncs: {}'.format(prob.dof_size, dofs.nfuncs))
    logger.info(f'nparams: {prob.dof_size}')
    logger.info(f'x0: {x0}')

    # Set up the list of parameter values to try
    if centered:
        nevals_jac = 2 * prob.dof_size
        xs = np.zeros((prob.dof_size, nevals_jac))
        for j in range(prob.dof_size):
            xs[:, 2 * j] = x0[:]  # I don't think I need np.copy(), but not 100% sure.
            xs[j, 2 * j] = x0[j] + eps
            xs[:, 2 * j + 1] = x0[:]
            xs[j, 2 * j + 1] = x0[j] - eps
    else:
        # 1-sided differences
        nevals_jac = prob.dof_size + 1
        xs = np.zeros((prob.dof_size, nevals_jac))
        xs[:, 0] = x0[:]
        for j in range(prob.dof_size):
            xs[:, j + 1] = x0[:]
            xs[j, j + 1] = x0[j] + eps

    # proc0_world will be responsible for detecting nvals, since
    #proc0_world always does at least 1 function evaluation. Other
    #procs cannot be trusted to evaluate nvals because they may
    #not have any function evals, in which case they never create
    #"evals", so the MPI reduce would fail.

    #evals = np.zeros((dofs.nfuncs, nevals_jac))
    evals = None
    if not mpi.proc0_world:
        # All procs other than proc0_world should initialize evals
        # before the nevals_jac loop, since they may not have any
        # evals.
        nvals = mpi.comm_leaders.bcast(prob.dof_size)
        evals = np.zeros((nvals, nevals_jac))
    # Do the hard work of evaluating the functions.
    for j in range(nevals_jac):
        # Handle only this group's share of the work:
        if np.mod(j, mpi.ngroups) == mpi.rank_leaders:
            mpi.mobilize_workers(CALCULATE_F)
            x = xs[:, j]
            mpi.comm_groups.bcast(x, root=0)
            prob.x = x

            try:
                f = prob.objective()
            except:
                logger.info("Exception caught during function evaluation")
                f = np.full(prob.dofs.nvals, 1.0e12)

            if evals is None and mpi.proc0_world:
                nvals = mpi.comm_leaders.bcast(prob.dof_size)
                evals = np.zeros((nvals, nevals_jac))

            evals[:, j] = f
            #evals[:, j] = np.array([f() for f in dofs.funcs])

    # Combine the results from all groups:
    evals = mpi.comm_leaders.reduce(evals, op=MPI.SUM, root=0)

    if not apart_at_start:
        mpi.stop_workers()

    # Only proc0_world will actually have the Jacobian.
    if not mpi.proc0_world:
        return (None, None, None)

    # Use the evals to form the Jacobian
    jac = np.zeros((prob.dof_size, prob.dof_size))
    if centered:
        for j in range(prob.dof_size):
            jac[:, j] = (evals[:, 2 * j] - evals[:, 2 * j + 1]) / (2 * eps)
    else:
        # 1-sided differences:
        for j in range(prob.dof_size):
            jac[:, j] = (evals[:, j + 1] - evals[:, 0]) / eps

    # Weird things may happen if we do not reset the state vector
    # to x0:
    prob.x = x0
    return jac, xs, evals


def least_squares_mpi_solve(prob: LeastSquaresProblem,
                            mpi: MpiPartition,
                            grad: bool = None,
                            **kwargs):
    """
    Solve a nonlinear-least-squares minimization problem using
    MPI. All MPI processes (including group leaders and workers)
    should call this function.

    Args:
        prob: Optimizable object defining the objective function(s) and
              parameter space.
        mpi: A MpiPartition object, storing the information about how
             the pool of MPI processes is divided into worker groups.
        grad: Whether to use a gradient-based optimization algorithm, as
              opposed to a gradient-free algorithm. If unspecified, a
              gradient-based algorithm will be used if ``prob`` has gradient
              information available, otherwise a gradient-free algorithm
              will be used by default. If you set ``grad=True`` for a problem
              in which gradient information is not available,
              finite-difference gradients will be used.
        kwargs: Any arguments to pass to
                `scipy.optimize.least_squares <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.least_squares.html>`_.
                For instance, you can supply ``max_nfev=100`` to set
                the maximum number of function evaluations (not counting
                finite-difference gradient evaluations) to 100. Or, you
                can supply ``method`` to choose the optimization algorithm.
    """
    if MPI is None:
        raise RuntimeError(
            "least_squares_mpi_solve requires the mpi4py package.")

    logger.info("Beginning solve.")
    #prob._init()
    #if grad is None:
    #    grad = prob.grad_avail

    x = np.copy(prob.x)  # For use in Bcast later.

    objective_file = None
    residuals_file = None
    datalog_started = False
    nevals = 0
    start_time = time()

    def _f_proc0(x):
        """
        This function is used for least_squares_mpi_solve.  It is similar
        to LeastSquaresProblem.f(), except this version is called only by
        proc 0 while workers are in the worker loop.
        """
        logger.debug("Entering _f_proc0")
        mpi.mobilize_workers(CALCULATE_F)
        # Send workers the state vector:
        mpi.comm_groups.bcast(x, root=0)
        logger.debug("Past bcast in _f_proc0")

        try:
            #f_unshifted = prob.dofs.f(x)
            unweighted_residuals = prob.unweighted_residuals(x)
            logger.debug(f"unweighted residuals:\n {unweighted_residuals}")
            residuals = prob.residuals()
            logger.debug(f"residuals:\n {residuals}")
        except:
            #f_unshifted = np.full(prob.dofs.nvals, 1.0e12)
            unweighted_residuals = np.full(prob.get_parent_return_fns_no(),
                                           1.0e12)
            residuals = np.full(prob.get_parent_return_fns_no(), 1.0e12)
            logger.info("Exception caught during function evaluation.")

        #f_shifted = prob.f_from_unshifted(f_unshifted)
        #objective_val = prob.objective_from_shifted_f(f_shifted)
        objective_val = prob.objective()

        nonlocal datalog_started, objective_file, residuals_file, nevals

        # Since the number of terms is not known until the first
        # evaluation of the objective function, we cannot write the
        # header of the output file until this first evaluation is
        # done.
        if not datalog_started:
            # Initialize log file
            datalog_started = True
            datestr = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
            objective_file = open(f"objective_{datestr}.dat", 'w')
            objective_file.write("Problem type:\nleast_squares")
            objective_file.write("function_evaluation,seconds")

            residuals_file = open(f"residuals_{datestr}.dat", 'w')
            residuals_file.write("Problem type:\nleast_squares")
            residuals_file.write("function_evaluation,seconds")

            for j in range(prob.dof_size):
                objective_file.write(f",x({j})")
            objective_file.write(",objective_function\n")

            for j in range(prob.dof_size):
                residuals_file.write(f",x({j})")
            residuals_file.write(",objective_function")
            for j in range(prob.dof_size):
                residuals_file.write(f",F({j})")
            residuals_file.write("\n")

        del_t = time() - start_time
        objective_file.write(f"{nevals:6d},{del_t:12.4e}")
        for xj in x:
            objective_file.write(f",{xj:24.16e}")
        objective_file.write(f",{objective_val:24.16e}\n")
        objective_file.flush()

        residuals_file.write(f"{nevals:6d},{del_t:12.4e}")
        for xj in x:
            residuals_file.write(f",{xj:24.16e}")
        residuals_file.write(f",{objective_val:24.16e}")
        for fj in unweighted_residuals:
            residuals_file.write(f",{fj:24.16e}")
        residuals_file.write("\n")
        residuals_file.flush()

        nevals += 1
        return residuals

    def _jac_proc0(x):
        """
        This function is used for least_squares_mpi_solve.  It is similar
        to LeastSquaresProblem.jac, except this version is called only by
        proc 0 while workers are in the worker loop.
        """
        return None
        # TODO: reimplement this after jac evaluation is implemented
        if prob.grad_avail:
            # proc0_world calling mobilize_workers will mobilize only group 0.
            mpi.mobilize_workers(CALCULATE_JAC)
            # Send workers the state vector:
            mpi.comm_groups.bcast(x, root=0)

            return prob.jac(x)

        else:
            # Evaluate Jacobian using fd_jac_mpi
            mpi.mobilize_leaders(CALCULATE_FD_JAC)
            # Send leaders the state vector:
            mpi.comm_leaders.bcast(x, root=0)

            jac, xs, evals = fd_jac_mpi(prob, mpi, x)

            # Write function evaluations to the files
            nonlocal datalog_started, objective_file, residuals_file, nevals
            nevals_jac = evals.shape[1]
            for j in range(nevals_jac):
                objective_val = prob.objective_from_unshifted_f(evals[:, j])

                del_t = time() - start_time
                objective_file.write(f"{nevals:6d},{del_t:12.4e}")
                for xj in xs[:, j]:
                    objective_file.write(f",{xj:24.16e}")
                objective_file.write(f",{objective_val:24.16e}")
                objective_file.write("\n")
                objective_file.flush()

                residuals_file.write(f"{nevals:6d},{del_t:12.4e}")
                for xj in xs[:, j]:
                    residuals_file.write(f",{xj:24.16e}")
                residuals_file.write(f",{objective_val:24.16e}")
                for fj in evals[:, j]:
                    residuals_file.write(f",{fj:24.16e}")
                residuals_file.write("\n")
                residuals_file.flush()

                nevals += 1

            return prob.scale_dofs_jac(jac)

    # Send group leaders and workers into their respective loops:
    leaders_action = lambda mpi2, data: _mpi_leaders_task(mpi, prob, data)
    workers_action = lambda mpi2, data: _mpi_workers_task(mpi, prob, data)
    mpi.apart(leaders_action, workers_action)

    if mpi.proc0_world:
        # proc0_world does this block, running the optimization.
        x0 = np.copy(prob.x)
        #print("x0:",x0)
        # Call scipy.optimize:
        if grad:
            logger.info("Using derivative free method despite derivatives given")
            #result = least_squares(_f_proc0, x0, verbose=2, jac=_jac_proc0, **kwargs)
            result = least_squares(_f_proc0, x0, verbose=2, **kwargs)
        else:
            logger.info("Using derivative-free method")
            result = least_squares(_f_proc0, x0, verbose=2, **kwargs)

        logger.info("Completed solve.")
        x = result.x

        objective_file.close()
        residuals_file.close()

    # Stop loops for workers and group leaders:
    mpi.together()

    datalog_started = False
    logger.info("Completed solve.")

    # Finally, make sure all procs get the optimal state vector.
    mpi.comm_world.Bcast(x)
    logger.debug(f'After Bcast, x={x}')
    # Set Parameters to their values for the optimum
    prob.x = x




import numpy as np
import logging

from .surface import Surface
from .surfacerzfourier import SurfaceRZFourier

logger = logging.getLogger(__name__)


class SurfaceGarabedian(Surface):
    """
    `SurfaceGarabedian` represents a toroidal surface for which the
    shape is parameterized using Garabedian's :math:`\Delta_{m,n}`
    coefficients.

    The present implementation assumes stellarator symmetry. Note that
    non-stellarator-symmetric surfaces require that the :math:`\Delta_{m,n}`
    coefficients be imaginary.
    """

    def __init__(self, nfp=1, mmax=1, mmin=0, nmax=0, nmin=None):
        if nmin is None:
            nmin = -nmax
        # Perform some validation.
        if mmax < mmin:
            raise ValueError("mmin must be >= mmax")
        if nmax < nmin:
            raise ValueError("nmin must be >= nmax")
        if mmax < 1:
            raise ValueError("mmax must be >= 1")
        if mmin > 0:
            raise ValueError("mmin must be <= 0")
        self.mmin = mmin
        self.mmax = mmax
        self.nmin = nmin
        self.nmax = nmax
        self.nfp = nfp
        self.stellsym = True
        self.allocate()
        self.recalculate = True
        self.recalculate_derivs = True

        # Initialize to an axisymmetric torus with major radius 1m and
        # minor radius 0.1m
        self.set_Delta(1, 0, 1.0)
        self.set_Delta(0, 0, 0.1)
        Surface.__init__(self)

    def __repr__(self):
        return "SurfaceGarabedian " + str(hex(id(self))) + " (nfp=" + \
            str(self.nfp) + ", mmin=" + str(self.mmin) + ", mmax=" + str(self.mmax) \
            + ", nmin=" + str(self.nmin) + ", nmax=" + str(self.nmax) \
            + ")"

    def allocate(self):
        """
        Create the array for the :math:`\Delta_{m,n}` coefficients.
        """
        logger.info("Allocating SurfaceGarabedian")
        self.mdim = self.mmax - self.mmin + 1
        self.ndim = self.nmax - self.nmin + 1
        myshape = (self.mdim, self.ndim)
        self.Delta = np.zeros(myshape)
        self.names = []
        for n in range(self.nmin, self.nmax + 1):
            for m in range(self.mmin, self.mmax + 1):
                self.names.append('Delta(' + str(m) + ',' + str(n) + ')')

    def get_Delta(self, m, n):
        """
        Return a particular :math:`\Delta_{m,n}` coefficient.
        """
        return self.Delta[m - self.mmin, n - self.nmin]

    def set_Delta(self, m, n, val):
        """
        Set a particular :math:`\Delta_{m,n}` coefficient.
        """
        self.Delta[m - self.mmin, n - self.nmin] = val
        self.recalculate = True
        self.recalculate_derivs = True

    def get_dofs(self):
        """
        Return a 1D numpy array with all the degrees of freedom.
        """
        num_dofs = (self.mmax - self.mmin + 1) * (self.nmax - self.nmin + 1)
        return np.reshape(self.Delta, (num_dofs,), order='F')

    def set_dofs(self, v):
        """
        Set the shape coefficients from a 1D list/array
        """

        n = len(self.get_dofs())
        if len(v) != n:
            raise ValueError('Input vector should have ' + str(n) + \
                             ' elements but instead has ' + str(len(v)))

        # Check whether any elements actually change:
        if np.all(np.abs(self.get_dofs() - np.array(v)) == 0):
            logger.info('set_dofs called, but no dofs actually changed')
            return

        logger.info('set_dofs called, and at least one dof changed')
        self.recalculate = True
        self.recalculate_derivs = True

        self.Delta = v.reshape((self.mmax - self.mmin + 1, self.nmax - self.nmin + 1), order='F')

    def fixed_range(self, mmin, mmax, nmin, nmax, fixed=True):
        """
        Set the 'fixed' property for a range of m and n values.

        All modes with m in the interval [mmin, mmax] and n in the
        interval [nmin, nmax] will have their fixed property set to
        the value of the 'fixed' parameter. Note that mmax and nmax
        are included (unlike the upper bound in python's range(min,
        max).)
        """
        for m in range(mmin, mmax + 1):
            for n in range(nmin, nmax + 1):
                self.set_fixed('Delta({},{})'.format(m, n), fixed)

    def to_RZFourier(self):
        """
        Return a SurfaceRZFourier object with the identical shape.

        For a derivation of the transformation here, see 
        https://terpconnect.umd.edu/~mattland/assets/notes/toroidal_surface_parameterizations.pdf
        """
        mpol = int(np.max((1, self.mmax - 1, 1 - self.mmin)))
        ntor = int(np.max((self.nmax, -self.nmin)))
        s = SurfaceRZFourier(nfp=self.nfp, stellsym=True, mpol=mpol, ntor=ntor)
        s.set_rc(0, 0, self.get_Delta(1, 0))
        for m in range(mpol + 1):
            nmin = -ntor
            if m == 0:
                nmin = 1
            for n in range(nmin, ntor + 1):
                Delta1 = 0
                Delta2 = 0
                if 1 - m >= self.mmin and -n >= self.nmin and -n <= self.nmax:
                    Delta1 = self.get_Delta(1 - m, -n)
                if 1 + m <= self.mmax and n >= self.nmin and n <= self.nmax:
                    Delta2 = self.get_Delta(1 + m, n)
                s.set_rc(m, n, Delta1 + Delta2)
                s.set_zs(m, n, Delta1 - Delta2)

        return s

    def area_volume(self):
        """
        Compute the surface area and the volume enclosed by the surface.
        """
        if self.recalculate:
            logger.info('Running calculation of area and volume')
        else:
            logger.info('area_volume called, but no need to recalculate')
            return

        self.recalculate = False

        # Delegate to the area and volume calculations of SurfaceRZFourier():
        s = self.to_RZFourier()
        self._area = s.area()
        self._volume = s.volume()

    def area(self):
        """
        Return the area of the surface.
        """
        self.area_volume()
        return self._area

    def volume(self):
        """
        Return the volume of the surface.
        """
        self.area_volume()
        return self._volume




from math import sqrt
import numpy as np
import simsoptpp as sopp
import logging
from simsopt.field.magneticfield import MagneticField
from simsopt.field.sampling import draw_uniform_on_curve, draw_uniform_on_surface
from simsopt.geo.surface import SurfaceClassifier
from simsopt.util.constants import ALPHA_PARTICLE_MASS, ALPHA_PARTICLE_CHARGE, FUSION_ALPHA_PARTICLE_ENERGY
from nptyping import NDArray, Float


logger = logging.getLogger(__name__)


def compute_gc_radius(m, vperp, q, absb):
    """
    Computes the gyro radius of a particle in a field with strenght ``absb```,
    that is ``r=m*vperp/(abs(q)*absb)``.
    """

    return m*vperp/(abs(q)*absb)


def gc_to_fullorbit_initial_guesses(field, xyz_inits, speed_pars, speed_total, m, q, eta=0):
    """
    Takes in guiding center positions ``xyz_inits`` as well as a parallel
    speeds ``speed_pars`` and total velocities ``speed_total`` to compute orbit
    positions for a full orbit calculation that matches the given guiding
    center. The phase angle can be controll via the `eta` parameter
    """

    nparticles = xyz_inits.shape[0]
    xyz_inits_full = np.zeros_like(xyz_inits)
    v_inits = np.zeros((nparticles, 3))
    rgs = np.zeros((nparticles, ))
    field.set_points(xyz_inits)
    Bs = field.B()
    AbsBs = field.AbsB()
    eB = Bs/AbsBs
    for i in range(nparticles):
        p1 = eB[i, :]
        p2 = np.asarray([0, 0, 1])
        p3 = -np.cross(p1, p2)
        p3 *= 1./np.linalg.norm(p3)
        q1 = p1
        q2 = p2 - np.sum(q1*p2)*q1
        q2 = q2/np.sum(q2**2)**0.5
        q3 = p3 - np.sum(q1*p3)*q1 - np.sum(q2*p3)*q2
        q3 = q3/np.sum(q3**2)**0.5
        speed_perp = np.sqrt(speed_total**2 - speed_pars[i]**2)
        rgs[i] = compute_gc_radius(m, speed_perp, q, AbsBs[i, 0])

        xyz_inits_full[i, :] = xyz_inits[i, :] + rgs[i] * np.sin(eta) * q2 + rgs[i] * np.cos(eta) * q3
        vperp = -speed_perp * np.cos(eta) * q2 + speed_perp * np.sin(eta) * q3
        v_inits[i, :] = speed_pars[i] * q1 + vperp
    return xyz_inits_full, v_inits, rgs


def parallel_loop_bounds(comm, n):
    """
    Split up an array [0, 1, ..., n-1] across an mpi communicator.  Example: n
    = 8, comm with size=2 will return (0, 4) on core 0, (4, 8) on core 1,
    meaning that the array is split up as [0, 1, 2, 3] + [4, 5, 6, 7].
    """

    if comm is None:
        return 0, n
    else:
        size = comm.size
        idxs = [i*n//size for i in range(size+1)]
        assert idxs[0] == 0
        assert idxs[-1] == n
        return idxs[comm.rank], idxs[comm.rank+1]


def trace_particles(field: MagneticField, xyz_inits: NDArray[Float],
                    parallel_speeds: NDArray[Float], tmax=1e-4,
                    mass=ALPHA_PARTICLE_MASS, charge=ALPHA_PARTICLE_CHARGE, Ekin=FUSION_ALPHA_PARTICLE_ENERGY,
                    tol=1e-9, comm=None, phis=[], stopping_criteria=[], mode='gc_vac', forget_exact_path=False):
    r"""
    Follow particles in a magnetic field.

    In the case of ``mod='full'`` we solve

    .. math::

        [\ddot x, \ddot y, \ddot z] = \frac{q}{m}  [\dot x, \dot y, \dot z] \times B

    in the case of ``mod='gc_vac'`` we solve the guiding center equations under
    the assumption :math:`\nabla p=0`, that is

    .. math::

        [\dot x, \dot y, \dot z] &= v_{||}\frac{B}{|B|} + \frac{m}{q|B|^3}  (0.5v_\perp^2 + v_{||}^2)  B\times \nabla(|B|)\\
        \dot v_{||}    &= -\mu  (B \cdot \nabla(|B|))

    where :math:`v_\perp = 2\mu|B|`. See equations (12) and (13) of
    [Guiding Center Motion, H.J. de Blank, https://doi.org/10.13182/FST04-A468].

    Args:
        field: The magnetic field :math:`B`.
        xyz_inits: A (nparticles, 3) array with the initial positions of the particles.
        parallel_speeds: A (nparticles, ) array containing the speed in direction of the B field
                         for each particle.
        tmax: integration time
        mass: particle mass in kg, defaults to the mass of an alpha particle
        charge: charge in Coulomb, defaults to the charge of an alpha particle
        Ekin: kinetic energy in Joule, defaults to 3.52MeV
        tol: tolerance for the adaptive ode solver
        comm: MPI communicator to parallelize over
        phis: list of angles in [0, 2pi] for which intersection with the plane
              corresponding to that phi should be computed
        stopping_criteria: list of stopping criteria, mostly used in
                           combination with the ``LevelsetStoppingCriterion``
                           accessed via :obj:`simsopt.field.tracing.SurfaceClassifier`.
        mode: how to trace the particles. options are
            `gc`: general guiding center equations,
            `gc_vac`: simplified guiding center equations for the case :math:`\nabla p=0`,
            `full`: full orbit calculation (slow!)
        forget_exact_path: return an empty list for the ``res_tys``. To be used when only res_phi_hits
                           is of interest and one wants to reduce memory usage.

    Returns: 2 element tuple containing
        - ``res_tys``:
            A list of numpy arrays (one for each particle) describing the
            solution over time. The numpy array is of shape (ntimesteps, M)
            with M depending on the ``mode``.  Each row contains the time and
            the state.  So for `mode='gc'` and `mode='gc_vac'` the state
            consists of the xyz position and the parallel speed, hence
            each row contains `[t, x, y, z, v_par]`.  For `mode='full'`, the
            state consists of position and velocity vector, i.e. each row
            contains `[t, x, y, z, vx, vy, vz]`.

        - ``res_phi_hits``:
            A list of numpy arrays (one for each particle) containing
            information on each time the particle hits one of the phi planes or
            one of the stopping criteria. Each row of the array contains
            `[time] + [idx] + state`, where `idx` tells us which of the `phis`
            or `stopping_criteria` was hit.  If `idx>=0`, then `phis[int(idx)]`
            was hit. If `idx<0`, then `stopping_criteria[int(-idx)-1]` was hit.
    """

    nparticles = xyz_inits.shape[0]
    assert xyz_inits.shape[0] == len(parallel_speeds)
    speed_par = parallel_speeds
    mode = mode.lower()
    assert mode in ['gc', 'gc_vac', 'full']
    m = mass
    speed_total = sqrt(2*Ekin/m)  # Ekin = 0.5 * m * v^2 <=> v = sqrt(2*Ekin/m)

    if mode == 'full':
        xyz_inits, v_inits, _ = gc_to_fullorbit_initial_guesses(field, xyz_inits, speed_par, speed_total, m, charge)
    res_tys = []
    res_phi_hits = []
    loss_ctr = 0
    first, last = parallel_loop_bounds(comm, nparticles)
    for i in range(first, last):
        if 'gc' in mode:
            res_ty, res_phi_hit = sopp.particle_guiding_center_tracing(
                field, xyz_inits[i, :],
                m, charge, speed_total, speed_par[i], tmax, tol,
                vacuum=(mode == 'gc_vac'), phis=phis, stopping_criteria=stopping_criteria)
        else:
            res_ty, res_phi_hit = sopp.particle_fullorbit_tracing(
                field, xyz_inits[i, :], v_inits[i, :],
                m, charge, tmax, tol, phis=phis, stopping_criteria=stopping_criteria)
        if not forget_exact_path:
            res_tys.append(np.asarray(res_ty))
        else:
            res_tys.append([])
        res_phi_hits.append(np.asarray(res_phi_hit))
        dtavg = res_ty[-1][0]/len(res_ty)
        logger.debug(f"{i+1:3d}/{nparticles}, t_final={res_ty[-1][0]}, average timestep {1000*dtavg:.10f}ms")
        if res_ty[-1][0] < tmax - 1e-15:
            loss_ctr += 1
    if comm is not None:
        loss_ctr = comm.allreduce(loss_ctr)
    if comm is not None:
        res_tys = [i for o in comm.allgather(res_tys) for i in o]
        res_phi_hits = [i for o in comm.allgather(res_phi_hits) for i in o]
    logger.debug(f'Particles lost {loss_ctr}/{nparticles}={(100*loss_ctr)//nparticles:d}%')
    return res_tys, res_phi_hits


def trace_particles_starting_on_curve(curve, field, nparticles, tmax=1e-4,
                                      mass=ALPHA_PARTICLE_MASS, charge=ALPHA_PARTICLE_CHARGE,
                                      Ekin=FUSION_ALPHA_PARTICLE_ENERGY,
                                      tol=1e-9, comm=None, seed=1, umin=-1, umax=+1,
                                      phis=[], stopping_criteria=[], mode='gc_vac', forget_exact_path=False):
    r"""
    Follows particles spawned at random locations on the magnetic axis with random pitch angle.
    See :mod:`simsopt.field.tracing.trace_particles` for the governing equations.

    Args:
        curve: The :mod:`simsopt.geo.curve.Curve` to spawn the particles on. Uses rejection sampling
               to sample points on the curve. *Warning*: assumes that the underlying
               quadrature points on the Curve are uniformly distributed.
        field: The magnetic field :math:`B`.
        nparticles: number of particles to follow.
        tmax: integration time
        mass: particle mass in kg, defaults to the mass of an alpha particle
        charge: charge in Coulomb, defaults to the charge of an alpha particle
        Ekin: kinetic energy in Joule, defaults to 3.52MeV
        tol: tolerance for the adaptive ode solver
        comm: MPI communicator to parallelize over
        seed: random seed
        umin: the parallel speed is defined as  ``v_par = u * speed_total``
            where  ``u`` is drawn uniformly in ``[umin, umax]``
        umax: see ``umin``
        phis: list of angles in [0, 2pi] for which intersection with the plane
              corresponding to that phi should be computed
        stopping_criteria: list of stopping criteria, mostly used in
                           combination with the ``LevelsetStoppingCriterion``
                           accessed via :obj:`simsopt.field.tracing.SurfaceClassifier`.
        mode: how to trace the particles. options are
            `gc`: general guiding center equations,
            `gc_vac`: simplified guiding center equations for the case :math:`\nabla p=0`,
            `full`: full orbit calculation (slow!)
        forget_exact_path: return an empty list for the ``res_tys``. To be used when only res_phi_hits
                           is of interest and one wants to reduce memory usage.

    Returns: see :mod:`simsopt.field.tracing.trace_particles`
    """
    m = mass
    speed_total = sqrt(2*Ekin/m)  # Ekin = 0.5 * m * v^2 <=> v = sqrt(2*Ekin/m)
    np.random.seed(seed)
    us = np.random.uniform(low=umin, high=umax, size=(nparticles, ))
    speed_par = us*speed_total
    xyz, _ = draw_uniform_on_curve(curve, nparticles, safetyfactor=10)
    return trace_particles(
        field, xyz, speed_par, tmax=tmax, mass=mass, charge=charge,
        Ekin=Ekin, tol=tol, comm=comm, phis=phis,
        stopping_criteria=stopping_criteria, mode=mode, forget_exact_path=forget_exact_path)


def trace_particles_starting_on_surface(surface, field, nparticles, tmax=1e-4,
                                        mass=ALPHA_PARTICLE_MASS, charge=ALPHA_PARTICLE_CHARGE,
                                        Ekin=FUSION_ALPHA_PARTICLE_ENERGY,
                                        tol=1e-9, comm=None, seed=1, umin=-1, umax=+1,
                                        phis=[], stopping_criteria=[], mode='gc_vac', forget_exact_path=False):
    r"""
    Follows particles spawned at random locations on the magnetic axis with random pitch angle.
    See :mod:`simsopt.field.tracing.trace_particles` for the governing equations.

    Args:
        surface: The :mod:`simsopt.geo.surface.Surface` to spawn the particles
                 on. Uses rejection sampling to sample points on the curve. *Warning*:
                 assumes that the underlying quadrature points on the Curve as uniformly
                 distributed.
        field: The magnetic field :math:`B`.
        nparticles: number of particles to follow.
        tmax: integration time
        mass: particle mass in kg, defaults to the mass of an alpha particle
        charge: charge in Coulomb, defaults to the charge of an alpha particle
        Ekin: kinetic energy in Joule, defaults to 3.52MeV
        tol: tolerance for the adaptive ode solver
        comm: MPI communicator to parallelize over
        seed: random seed
        umin: the parallel speed is defined as  ``v_par = u * speed_total``
            where  ``u`` is drawn uniformly in ``[umin, umax]``
        umax: see ``umin``
        phis: list of angles in [0, 2pi] for which intersection with the plane
              corresponding to that phi should be computed
        stopping_criteria: list of stopping criteria, mostly used in
                           combination with the ``LevelsetStoppingCriterion``
                           accessed via :obj:`simsopt.field.tracing.SurfaceClassifier`.
        mode: how to trace the particles. options are
            `gc`: general guiding center equations,
            `gc_vac`: simplified guiding center equations for the case :math:`\nabla p=0`,
            `full`: full orbit calculation (slow!)
        forget_exact_path: return an empty list for the ``res_tys``. To be used when only res_phi_hits
                           is of interest and one wants to reduce memory usage.

    Returns: see :mod:`simsopt.field.tracing.trace_particles`
    """
    m = mass
    speed_total = sqrt(2*Ekin/m)  # Ekin = 0.5 * m * v^2 <=> v = sqrt(2*Ekin/m)
    np.random.seed(seed)
    us = np.random.uniform(low=umin, high=umax, size=(nparticles, ))
    speed_par = us*speed_total
    xyz, _ = draw_uniform_on_surface(surface, nparticles, safetyfactor=10)
    return trace_particles(
        field, xyz, speed_par, tmax=tmax, mass=mass, charge=charge,
        Ekin=Ekin, tol=tol, comm=comm, phis=phis,
        stopping_criteria=stopping_criteria, mode=mode, forget_exact_path=forget_exact_path)


def compute_fieldlines(field, R0, Z0, tmax=200, tol=1e-7, phis=[], stopping_criteria=[], comm=None):
    r"""
    Compute magnetic field lines by solving

    .. math::

        [\dot x, \dot y, \dot z] = B(x, y, z)

    Args:
        field: the magnetic field :math:`B`
        R0: list of radial components of initial points
        Z0: list of vertical components of initial points
        tmax: for how long to trace. will do roughly |B|*tmax/(2*pi*r0) revolutions of the device
        tol: tolerance for the adaptive ode solver
        phis: list of angles in [0, 2pi] for which intersection with the plane
              corresponding to that phi should be computed
        stopping_criteria: list of stopping criteria, mostly used in
                           combination with the ``LevelsetStoppingCriterion``
                           accessed via :obj:`simsopt.field.tracing.SurfaceClassifier`.

    Returns: 2 element tuple containing
        - ``res_tys``:
            A list of numpy arrays (one for each particle) describing the
            solution over time. The numpy array is of shape (ntimesteps, 4).
            Each row contains the time and
            the position, i.e.`[t, x, y, z]`.
        - ``res_phi_hits``:
            A list of numpy arrays (one for each particle) containing
            information on each time the particle hits one of the phi planes or
            one of the stopping criteria. Each row of the array contains
            `[time, idx, x, y, z]`, where `idx` tells us which of the `phis`
            or `stopping_criteria` was hit.  If `idx>=0`, then `phis[int(idx)]`
            was hit. If `idx<0`, then `stopping_criteria[int(-idx)-1]` was hit.
    """
    assert len(R0) == len(Z0)
    nlines = len(R0)
    xyz_inits = np.zeros((nlines, 3))
    xyz_inits[:, 0] = np.asarray(R0)
    xyz_inits[:, 2] = np.asarray(Z0)
    res_tys = []
    res_phi_hits = []
    first, last = parallel_loop_bounds(comm, nlines)
    for i in range(first, last):
        res_ty, res_phi_hit = sopp.fieldline_tracing(
            field, xyz_inits[i, :],
            tmax, tol, phis=phis, stopping_criteria=stopping_criteria)
        res_tys.append(np.asarray(res_ty))
        res_phi_hits.append(np.asarray(res_phi_hit))
        dtavg = res_ty[-1][0]/len(res_ty)
        logger.debug(f"{i+1:3d}/{nlines}, t_final={res_ty[-1][0]}, average timestep {dtavg:.10f}s")
    if comm is not None:
        res_tys = [i for o in comm.allgather(res_tys) for i in o]
        res_phi_hits = [i for o in comm.allgather(res_phi_hits) for i in o]
    return res_tys, res_phi_hits


def particles_to_vtk(res_tys, filename):
    """
    Export particle tracing or field lines to a vtk file.
    Expects that the xyz positions can be obtained by ``xyz[:, 1:4]``.
    """
    from pyevtk.hl import polyLinesToVTK
    x = np.concatenate([xyz[:, 1] for xyz in res_tys])
    y = np.concatenate([xyz[:, 2] for xyz in res_tys])
    z = np.concatenate([xyz[:, 3] for xyz in res_tys])
    ppl = np.asarray([xyz.shape[0] for xyz in res_tys])
    data = np.concatenate([i*np.ones((res_tys[i].shape[0], )) for i in range(len(res_tys))])
    polyLinesToVTK(filename, x, y, z, pointsPerLine=ppl, pointData={'idx': data})


class LevelsetStoppingCriterion(sopp.LevelsetStoppingCriterion):
    r"""
    Based on a scalar function :math:`f:R^3\to R`, this criterion checks whether
    :math:`f(x, y, z) < 0` and stops the iteration once this is true.

    The idea is to use this for example with signed distance functions to a surface.
    """

    def __init__(self, classifier):
        assert isinstance(classifier, SurfaceClassifier) \
            or isinstance(classifier, sopp.RegularGridInterpolant3D)
        if isinstance(classifier, SurfaceClassifier):
            sopp.LevelsetStoppingCriterion.__init__(self, classifier.dist)
        else:
            sopp.LevelsetStoppingCriterion.__init__(self, classifier)


class IterationStoppingCriterion(sopp.IterationStoppingCriterion):
    """
    Stop the iteration once the maximum number of iterations is reached.
    """
    pass


def plot_poincare_data(fieldlines_phi_hits, phis, filename, mark_lost=False, aspect='equal', dpi=300):
    """
    Create a poincare plot. Usage:

    .. code-block::

        phis = np.linspace(0, 2*np.pi/nfp, nphis, endpoint=False)
        res_tys, res_phi_hits = compute_fieldlines(
            bsh, R0, Z0, tmax=1000, phis=phis, stopping_criteria=[])
        plot_poincare_data(res_phi_hits, phis, '/tmp/fieldlines.png')

    Requires matplotlib to be installed.

    """
    import matplotlib.pyplot as plt
    from math import ceil, sqrt
    nrowcol = ceil(sqrt(len(phis)))
    fig, axs = plt.subplots(nrowcol, nrowcol, figsize=(16, 10))
    color = None
    for i in range(len(phis)):
        row = i//nrowcol
        col = i % nrowcol
        axs[row, col].set_title(f"$\\phi = {phis[i]/np.pi:.3f}\\pi$ ", loc='right', y=0.0)
        axs[row, col].set_xlabel("$r$")
        axs[row, col].set_ylabel("$z$")
        axs[row, col].set_aspect(aspect)
        axs[row, col].tick_params(direction="in")
        for j in range(len(fieldlines_phi_hits)):
            lost = fieldlines_phi_hits[j][-1, 1] < 0
            if mark_lost:
                color = 'r' if lost else 'g'
            data_this_phi = fieldlines_phi_hits[j][np.where(fieldlines_phi_hits[j][:, 1] == i)[0], :]
            if data_this_phi.size == 0:
                continue
            r = np.sqrt(data_this_phi[:, 2]**2+data_this_phi[:, 3]**2)
            axs[row, col].scatter(r, data_this_phi[:, 4], marker='o', s=0.2, linewidths=0, c=color)
    plt.tight_layout()
    plt.savefig(filename, dpi=dpi)
    plt.close()



from jax.config import config
config.update("jax_enable_x64", True)
#from .magneticaxis import *
#from .fouriercurve import *
#from .objectives import *
from .config import *
#from .biotsavart import *
#from .curve import RotatedCurve



# ===================ATTENTION=================================================
# Don't abuse this file by importing all variables from all modules to top-level.
# Import only the important classes that should be at top-level.
# Follow the same logic in the sub-packages.
# ===================END ATTENTION=============================================

# Two ways of achieving the above-mentioned objective
# Use "from xyz import XYZ" style
# Define __all__ dunder at module and subpackage level. Then you could do 
# "from xyz import *".  If xyz[.py] contains __all__ = ['XYZ'], only XYZ is 
# imported

try:
    from importlib import metadata
except ImportError:
    # Running on pre-3.8 Python; use importlib-metadata package
    import importlib_metadata as metadata

__version__ = metadata.version('simsopt')

from ._core import make_optimizable
from .objectives import LeastSquaresTerm, LeastSquaresProblem
from .solve import least_squares_serial_solve
from .util import initialize_logging

#__all__ = ['LeastSquaresProblem', 'LeastSquaresTerm']



from releash import *

gitpush = ReleaseTargetGitPush('upstream', 'master')

xtl = add_package(path=".", name="xtl")
version_xtl = VersionSourceAndTargetHpp(xtl, '{path}/include/xtl/xtl_config.hpp', prefix='XTL_VERSION_')
gittag_xtl = ReleaseTargetGitTagVersion(version_source=version_xtl, prefix='', annotate=True)

xtl.version_source = version_xtl
xtl.version_targets.append(version_xtl)

xtl.release_targets.append(gittag_xtl)
xtl.release_targets.append(gitpush)

source_tarball_filename = 'https://github.com/xtensor-stack/xtl/archive/{version}.tar.gz'.format(version=version_xtl)
xtl.release_targets.append(ReleaseTargetCondaForge(xtl, '../xtl-feedstock', source_tarball_filename=source_tarball_filename))




# coding: utf-8
# Copyright (c) HiddenSymmetries Development Team.
# Distributed under the terms of the LGPL License

"""
This module provides a class that handles the SPEC equilibrium code.
"""

import logging
from typing import Union
import os.path
import traceback

import numpy as np

logger = logging.getLogger(__name__)

try:
    from mpi4py import MPI
except ImportError as e:
    MPI = None 
    logger.warning(str(e))

# spec_found = True
try:
    import spec
except ImportError as e:
    spec = None
    logger.warning(str(e))
    # spec_found = False

# py_spec_found = True
try:
    import py_spec
except ImportError as e:
    py_spec = None
    logger.warning(str(e))
    # py_spec_found = False

# pyoculus_found = True
try:
    import pyoculus
except ImportError as e:
    pyoculus = None
    logger.warning(str(e))
    # pyoculus_found = False

from .._core.optimizable import Optimizable
from .._core.util import ObjectiveFailure
from ..geo.surfacerzfourier import SurfaceRZFourier
from ..util.dev import SimsoptRequires
if MPI is not None:
    from ..util.mpi import MpiPartition
else:
    MpiPartition = None
#from ..util.mpi import MpiPartition


@SimsoptRequires(MPI is not None, "mpi4py needs to be installed for running SPEC")
class Spec(Optimizable):
    """
    This class represents the SPEC equilibrium code.

    Philosophy regarding mpol and ntor: The Spec object keeps track of
    mpol and ntor values that are independent of those for the
    boundary Surface object. If the Surface object has different
    mpol/ntor values, the Surface's rbc/zbs arrays are first copied,
    then truncated or expanded to fit the mpol/ntor values of the Spec
    object to before Spec is run. Therefore, you may sometimes need to
    manually change the mpol and ntor values for the Spec object.

    The default behavior is that all  output files will be
    deleted except for the first and most recent iteration on worker
    group 0. If you wish to keep all the output files, you can set
    ``keep_all_files = True``. If you want to save the output files
    for a certain intermediate iteration, you can set the
    ``files_to_delete`` attribute to ``[]`` after that run of SPEC.

    Args:
        filename: SPEC input file to use for initialization. It should end
          in ``.sp``. Or, if None, default values will be used.
        mpi: A :obj:`simsopt.util.mpi.MpiPartition` instance, from which
          the worker groups will be used for SPEC calculations. If ``None``,
          each MPI process will run SPEC independently.
        verbose: Whether to print SPEC output to stdout.
        keep_all_files: If ``False``, all output files will be deleted
          except for the first and most recent ones from worker group 0. If 
          ``True``, all output files will be kept.
    """

    def __init__(self,
                 filename: Union[str, None] = None,
                 mpi: Union[MpiPartition, None] = None,
                 verbose: bool = True,
                 keep_all_files: bool = False):

        #if not spec_found:
        if spec is None:
            raise RuntimeError(
                "Using Spec requires spec python wrapper to be installed.")

        #if not py_spec_found:
        if py_spec is None:
            raise RuntimeError(
                "Using Spec requires py_spec to be installed.")

        self.lib = spec
        # For the most commonly accessed fortran modules, provide a
        # shorthand so ".lib" is not needed:
        modules = [
            "inputlist",
            "allglobal",
        ]
        for key in modules:
            setattr(self, key, getattr(spec, key))

        self.verbose = verbose
        # mute screen output if necessary
        # TODO: relies on /dev/null being accessible (Windows!)
        if not self.verbose:
            self.lib.fileunits.mute(1)

        # python wrapper does not need to write files along the run
        #self.lib.allglobal.skip_write = True

        # If mpi is not specified, use a single worker group:
        if mpi is None:
            self.mpi = MpiPartition(ngroups=1)
        else:
            self.mpi = mpi
        # SPEC will use the "groups" communicator from the MpiPartition:
        self.lib.allglobal.set_mpi_comm(self.mpi.comm_groups.py2f())

        if filename is None:
            # Read default input file, which should be in the same
            # directory as this file:
            filename = os.path.join(os.path.dirname(__file__), 'defaults.sp')
            logger.info("Initializing a SPEC object from defaults in " \
                        + filename)
        else:
            if not filename.endswith('.sp'):
                filename = filename + '.sp'
            logger.info("Initializing a SPEC object from file: " + filename)

        self.init(filename)
        self.extension = filename[:-3]
        self.keep_all_files = keep_all_files
        self.files_to_delete = []

        # Create a surface object for the boundary:
        si = spec.inputlist  # Shorthand
        stellsym = bool(si.istellsym)
        print("In __init__, si.istellsym=", si.istellsym, " stellsym=", stellsym)
        self.boundary = SurfaceRZFourier(nfp=si.nfp,
                                         stellsym=stellsym,
                                         mpol=si.mpol,
                                         ntor=si.ntor)

        # Transfer the boundary shape from fortran to the boundary
        # surface object:
        for m in range(si.mpol + 1):
            for n in range(-si.ntor, si.ntor + 1):
                self.boundary.rc[m, n + si.ntor] = si.rbc[n + si.mntor, m + si.mmpol]
                self.boundary.zs[m, n + si.ntor] = si.zbs[n + si.mntor, m + si.mmpol]
                if not stellsym:
                    self.boundary.rs[m, n + si.ntor] = si.rbs[n + si.mntor, m + si.mmpol]
                    self.boundary.zc[m, n + si.ntor] = si.zbc[n + si.mntor, m + si.mmpol]

        self.depends_on = ["boundary"]
        self.need_to_run_code = True
        self.counter = -1

        # By default, all dofs owned by SPEC directly, as opposed to
        # dofs owned by the boundary surface object, are fixed.
        self.fixed = np.full(len(self.get_dofs()), True)
        self.names = ['phiedge', 'curtor']

    def get_dofs(self):
        return np.array([self.inputlist.phiedge,
                         self.inputlist.curtor])

    def set_dofs(self, x):
        self.need_to_run_code = True
        self.inputlist.phiedge = x[0]
        self.inputlist.curtor = x[1]

    def init(self, filename: str):
        """
        Initialize SPEC fortran state from an input file.

        Args:
            filename: Name of the file to load. It should end in ``.sp``.
        """
        logger.debug("Entering init")
        if self.mpi.proc0_groups:
            spec.inputlist.initialize_inputs()
            logger.debug("Done with initialize_inputs")
            self.extension = filename[:-3]  # Remove the ".sp"
            spec.allglobal.ext = self.extension
            spec.allglobal.read_inputlists_from_file()
            logger.debug("Done with read_inputlists_from_file")
            spec.allglobal.check_inputs()

        logger.debug('About to call broadcast_inputs')
        spec.allglobal.broadcast_inputs()
        logger.debug('About to call preset')
        spec.preset()
        logger.debug("Done with init")

    def run(self):
        """
        Run SPEC, if needed.
        """
        if not self.need_to_run_code:
            logger.info("run() called but no need to re-run SPEC.")
            return
        logger.info("Preparing to run SPEC.")
        self.counter += 1

        si = self.inputlist  # Shorthand

        # nfp must be consistent between the surface and SPEC. The surface's value trumps.
        si.nfp = self.boundary.nfp
        si.istellsym = int(self.boundary.stellsym)

        # Convert boundary to RZFourier if needed:
        boundary_RZFourier = self.boundary.to_RZFourier()

        # Transfer boundary data to fortran:
        si.rbc[:, :] = 0.0
        si.zbs[:, :] = 0.0
        si.rbs[:, :] = 0.0
        si.zbc[:, :] = 0.0
        mpol_capped = np.min([boundary_RZFourier.mpol, si.mmpol])
        ntor_capped = np.min([boundary_RZFourier.ntor, si.mntor])
        stellsym = bool(si.istellsym)
        print("In run, si.istellsym=", si.istellsym, " stellsym=", stellsym)
        for m in range(mpol_capped + 1):
            for n in range(-ntor_capped, ntor_capped + 1):
                si.rbc[n + si.mntor, m + si.mmpol] = boundary_RZFourier.get_rc(m, n)
                si.zbs[n + si.mntor, m + si.mmpol] = boundary_RZFourier.get_zs(m, n)
                if not stellsym:
                    si.rbs[n + si.mntor, m + si.mmpol] = boundary_RZFourier.get_rs(m, n)
                    si.zbc[n + si.mntor, m + si.mmpol] = boundary_RZFourier.get_zc(m, n)

        # Set the coordinate axis using the lrzaxis=2 feature:
        si.lrzaxis = 2
        # lrzaxis=2 only seems to work if the axis is not already set
        si.ras[:] = 0.0
        si.rac[:] = 0.0
        si.zas[:] = 0.0
        si.zac[:] = 0.0

        # Another possible way to initialize the coordinate axis: use
        # the m=0 modes of the boundary.
        # m = 0
        # for n in range(2):
        #     si.rac[n] = si.rbc[n + si.mntor, m + si.mmpol]
        #     si.zas[n] = si.zbs[n + si.mntor, m + si.mmpol]

        filename = self.extension + '_{:03}_{:06}'.format(self.mpi.group, self.counter)
        logger.info("Running SPEC using filename " + filename)
        self.allglobal.ext = filename
        try:
            # Here is where we actually run SPEC:
            if self.mpi.proc0_groups:
                logger.debug('About to call check_inputs')
                spec.allglobal.check_inputs()
            logger.debug('About to call broadcast_inputs')
            spec.allglobal.broadcast_inputs()
            logger.debug('About to call preset')
            spec.preset()
            logger.debug(f'About to call init_outfile')
            spec.sphdf5.init_outfile()
            logger.debug('About to call mirror_input_to_outfile')
            spec.sphdf5.mirror_input_to_outfile()
            if self.mpi.proc0_groups:
                logger.debug('About to call wrtend')
                spec.allglobal.wrtend()
            logger.debug('About to call init_convergence_output')
            spec.sphdf5.init_convergence_output()
            logger.debug(f'About to call spec')
            spec.spec()
            logger.debug('About to call diagnostics')
            spec.final_diagnostics()
            logger.debug('About to call write_grid')
            spec.sphdf5.write_grid()
            if self.mpi.proc0_groups:
                logger.debug('About to call wrtend')
                spec.allglobal.wrtend()
            logger.debug('About to call hdfint')
            spec.sphdf5.hdfint()
            logger.debug('About to call finish_outfile')
            spec.sphdf5.finish_outfile()
            logger.debug('About to call ending')
            spec.ending()

        except:
            if self.verbose:
                traceback.print_exc()
            raise ObjectiveFailure("SPEC did not run successfully.")

        logger.info("SPEC run complete.")
        # Barrier so workers do not try to read the .h5 file before it is finished:
        self.mpi.comm_groups.Barrier()

        try:
            self.results = py_spec.SPECout(filename + '.sp.h5')
        except:
            if self.verbose:
                traceback.print_exc()
            raise ObjectiveFailure("Unable to read results following SPEC execution")

        logger.info("Successfully loaded SPEC results.")
        self.need_to_run_code = False

        # Group leaders handle deletion of files:
        if self.mpi.proc0_groups:

            # If the worker group is not 0, delete all wout files, unless
            # keep_all_files is True:
            if (not self.keep_all_files) and (self.mpi.group > 0):
                os.remove(filename + '.sp.h5')
                os.remove(filename + '.sp.end')

            # Delete the previous output file, if desired:
            for file_to_delete in self.files_to_delete:
                os.remove(file_to_delete)
            self.files_to_delete = []

            # Record the latest output file to delete if we run again:
            if (self.mpi.group == 0) and (self.counter > 0) and (not self.keep_all_files):
                self.files_to_delete.append(filename + '.sp.h5')
                self.files_to_delete.append(filename + '.sp.end')

    def volume(self):
        """
        Return the volume inside the boundary flux surface.
        """
        self.run()
        return self.results.output.volume * self.results.input.physics.Nfp

    def iota(self):
        """
        Return the rotational transform in the middle of the volume.
        """
        self.run()
        return self.results.transform.fiota[1, 0]


class Residue(Optimizable):
    """
    Greene's residue, evaluated from a Spec equilibrum
    """

    def __init__(self, spec, pp, qq, vol=1, theta=0, s_guess=None, s_min=-1.0,
                 s_max=1.0, rtol=1e-9):
        """
        spec: a Spec object
        pp, qq: Numerator and denominator for the resonant iota = pp / qq
        vol: Index of the Spec volume to consider
        theta: Spec's theta coordinate at the periodic field line
        s_guess: Guess for the value of Spec's s coordinate at the periodic
                field line
        s_min, s_max: bounds on s for the search
        rtol: the relative tolerance of the integrator
        """
        # if not spec_found:
        if spec is None:
            raise RuntimeError(
                "Residue requires py_spec package to be installed.")
        # if not pyoculus_found:
        if pyoculus is None:
            raise RuntimeError(
                "Residue requires pyoculus package to be installed.")

        self.spec = spec
        self.pp = pp
        self.qq = qq
        self.vol = vol
        self.theta = theta
        self.rtol = rtol
        if s_guess is None:
            self.s_guess = 0.0
        else:
            self.s_guess = s_guess
        self.s_min = s_min
        self.s_max = s_max
        self.depends_on = ['spec']
        self.need_to_run_code = True
        self.fixed_point = None
        # We may at some point want to allow Residue to use a
        # different MpiPartition than the Spec object it is attached
        # to, but for now we'll use the same MpiPartition for
        # simplicity.
        self.mpi = spec.mpi

    def J(self):
        """
        Run Spec if needed, find the periodic field line, and return the residue
        """
        if not self.mpi.proc0_groups:
            logger.info("This proc is skipping Residue.J() since it is not a group leader.")
            return

        if self.need_to_run_code:
            self.spec.run()
            specb = pyoculus.problems.SPECBfield(self.spec.results, self.vol)
            # Set nrestart=0 because otherwise the random guesses in
            # pyoculus can cause examples/tests to be
            # non-reproducible.
            fp = pyoculus.solvers.FixedPoint(specb, {'theta': self.theta, 'nrestart': 0},
                                             integrator_params={'rtol': self.rtol})
            self.fixed_point = fp.compute(self.s_guess,
                                          sbegin=self.s_min,
                                          send=self.s_max,
                                          pp=self.pp, qq=self.qq)
            self.need_to_run_code = False

        if self.fixed_point is None:
            raise ObjectiveFailure("Residue calculation failed")

        return self.fixed_point.GreenesResidue

    def get_dofs(self):
        return np.array([])

    def set_dofs(self, x):
        self.need_to_run_code = True



# coding: utf-8
# Copyright (c) HiddenSymmetries Development Team.
# Distributed under the terms of the LGPL License

"""
This module provides a few minimal optimizable objects, each
representing a function. These functions are mostly used for testing.
"""

import numpy as np

from .._core.optimizable import Optimizable
from .._core.util import ObjectiveFailure


class Identity(Optimizable):
    """
    This class represents a term in an objective function which is just
    the identity. It has one degree of freedom, and the output of the function
    is equal to this degree of freedom.
    """

    def __init__(self, x=0.0):
        self.x = x
        self.dx = np.array([1.0])
        self.fixed = np.full(1, False)
        self.names = ['x']

    def J(self):
        return self.x

    def dJ(self):
        return np.array([1.0])

    @property
    def f(self):
        """
        Same as the function J(), but a property instead of a function.
        """
        return self.x

    @property
    def df(self):
        """
        Same as the function dJ(), but a property instead of a function.
        """
        return np.array([1.0])

    def get_dofs(self):
        return np.array([self.x])

    def set_dofs(self, xin):
        self.x = xin[0]


class Adder(Optimizable):
    """This class defines a minimal object that can be optimized. It has
    n degrees of freedom, and has a function that just returns the sum
    of these dofs. This class is used for testing.
    """

    def __init__(self, n=3):
        self.n = n
        self.x = np.zeros(n)
        self.fixed = np.full(n, False)        

    def J(self):
        """
        Returns the sum of the degrees of freedom.
        """
        return np.sum(self.x)

    def dJ(self):
        return np.ones(self.n)

    @property
    def f(self):
        """
        Same as the function J(), but a property instead of a function.
        """
        return self.J()

    @property
    def df(self):
        """
        Same as the function dJ(), but a property instead of a function.
        """
        return np.ones(self.n)

    def get_dofs(self):
        return self.x

    def set_dofs(self, xin):
        self.x = np.array(xin)


class Rosenbrock(Optimizable):
    """
    This class defines a minimal object that can be optimized.
    """

    def __init__(self, b=100.0, x=0.0, y=0.0):
        self._sqrtb = np.sqrt(b)
        self.names = ['x', 'y']
        self._x = x
        self._y = y
        self.fixed = np.full(2, False)        

    def term1(self):
        """
        Returns the first of the two quantities that is squared and summed.
        """
        return self._x - 1

    def term2(self):
        """
        Returns the second of the two quantities that is squared and summed.
        """
        return (self._x * self._x - self._y) / self._sqrtb

    def dterm1(self):
        """
        Returns the gradient of term1
        """
        return np.array([1.0, 0.0])

    def dterm2(self):
        """
        Returns the gradient of term2
        """
        return np.array([2 * self._x, -1.0]) / self._sqrtb

    @property
    def term1prop(self):
        """
        Same as term1, but a property rather than a callable function.
        """
        return self.term1()

    @property
    def term2prop(self):
        """
        Same as term2, but a property rather than a callable function.
        """
        return self.term2()

    @property
    def dterm1prop(self):
        """
        Same as dterm1, but a property rather than a callable function.
        """
        return self.dterm1()

    @property
    def dterm2prop(self):
        """
        Same as dterm2, but a property rather than a callable function.
        """
        return self.dterm2()

    def f(self):
        """
        Returns the total function, squaring and summing the two terms.
        """
        t1 = self.term1()
        t2 = self.term2()
        return t1 * t1 + t2 * t2

    def terms(self):
        """
        Returns term1 and term2 together as a 2-element numpy vector.
        """
        return np.array([self.term1(), self.term2()])

    def dterms(self):
        """
        Returns the 2x2 Jacobian for term1 and term2.
        """
        return np.array([[1.0, 0.0],
                         [2 * self._x / self._sqrtb, -1.0 / self._sqrtb]])

    def get_dofs(self):
        return np.array([self._x, self._y])

    def set_dofs(self, xin):
        self._x = xin[0]
        self._y = xin[1]


class RosenbrockWithFailures(Rosenbrock):
    """
    This class is similar to the Rosenbrock class, except that it
    fails (raising ObjectiveFailure) at regular intervals.  This is
    useful for testing that the simsopt infrastructure handles
    failures in the expected way.
    """

    def __init__(self, *args, fail_interval=8, **kwargs):
        self.nevals = 0
        self.fail_interval = fail_interval
        super().__init__(*args, **kwargs)

    def term1(self):
        self.nevals += 1
        if np.mod(self.nevals, self.fail_interval) == 0:
            raise ObjectiveFailure("Planned failure")

        return super().term1()


class TestObject1(Optimizable):
    """
    This is an optimizable object used for testing. It depends on two
    sub-objects, both of type Adder.
    """

    def __init__(self, val):
        self.val = val
        self.names = ['val']
        self.fixed = np.array([False])
        self.adder1 = Adder(3)
        self.adder2 = Adder(2)
        self.depends_on = ['adder1', 'adder2']

    def set_dofs(self, x):
        self.val = x[0]

    def get_dofs(self):
        return np.array([self.val])

    def J(self):
        return (self.val + 2 * self.adder1.J()) / (10.0 + self.adder2.J())

    def dJ(self):
        v = self.val
        a1 = self.adder1.J()
        a2 = self.adder2.J()
        # J = (v + 2 * a1) / (10 + a2)
        return np.concatenate((np.array([1.0 / (10.0 + a2)]),
                               np.full(self.adder1.n, 2.0 / (10.0 + a2)),
                               np.full(self.adder2.n, -(v + 2 * a1) / ((10.0 + a2) ** 2))))

    @property
    def f(self):
        """
        Same as J() but a property instead of a function.
        """
        return self.J()

    @property
    def df(self):
        """
        Same as dJ() but a property instead of a function.
        """
        return self.dJ()


class TestObject2(Optimizable):
    """
    This is an optimizable object used for testing. It depends on two
    sub-objects, both of type Adder.
    """

    def __init__(self, val1, val2):
        self.val1 = val1
        self.val2 = val2
        self.names = ['val1', 'val2']
        self.fixed = np.array([False, False])
        self.t = TestObject1(0.0)
        self.adder = Adder(2)
        self.depends_on = ['t', 'adder']

    def set_dofs(self, x):
        self.val1 = x[0]
        self.val2 = x[1]

    def get_dofs(self):
        return np.array([self.val1, self.val2])

    def J(self):
        v1 = self.val1
        v2 = self.val2
        t = self.t.J()
        a = self.adder.J()
        return v1 + a * np.cos(v2 + t)

    def dJ(self):
        v1 = self.val1
        v2 = self.val2
        a = self.adder.J()
        t = self.t.J()
        cosat = np.cos(v2 + t)
        sinat = np.sin(v2 + t)
        # Order of terms in the gradient: v1, v2, t, a
        return np.concatenate((np.array([1.0, -a * sinat]),
                               -a * sinat * self.t.dJ(),
                               cosat * self.adder.dJ()))

    @property
    def f(self):
        """
        Same as J() but a property instead of a function.
        """
        return self.J()

    @property
    def df(self):
        """
        Same as dJ() but a property instead of a function.
        """
        return self.dJ()


class Affine(Optimizable):
    """
    This class represents a random affine (i.e. linear plus constant)
    transformation from R^n to R^m.
    """

    def __init__(self, nparams, nvals):
        """
        nparams = number of independent variables.
        nvals = number of dependent variables.
        """
        self.nparams = nparams
        self.nvals = nvals
        self.A = (np.random.rand(nvals, nparams) - 0.5) * 4
        self.B = (np.random.rand(nvals) - 0.5) * 4
        self.x = np.zeros(nparams)

    def get_dofs(self):
        return self.x

    def set_dofs(self, x):
        self.x = x

    def J(self):
        return np.matmul(self.A, self.x) + self.B

    def dJ(self):
        return self.A


class Failer(Optimizable):
    """
    This class is used for testing failures of the objective
    function. This function always returns a vector with entries all
    1.0, except that ObjectiveFailure will be raised on a specified
    evaluation.

    Args:
        nparams: Number of input values.
        nvals: Number of entries in the return vector.
        fail_index: Which function evaluation to fail on.
    """

    def __init__(self,
                 nparams: int = 2,
                 nvals: int = 3,
                 fail_index: int = 2):
        self.nparams = nparams
        self.nvals = nvals
        self.fail_index = fail_index
        self.nevals = 0
        self.x = np.zeros(self.nparams)

    def J(self):
        self.nevals += 1
        if self.nevals == self.fail_index:
            raise ObjectiveFailure("nevals == fail_index")
        else:
            return np.full(self.nvals, 1.0)

    def get_dofs(self):
        return self.x

    def set_dofs(self, x):
        self.x = x


class Beale(Optimizable):
    """
    This is a test function which does not supply derivatives. It is
    taken from
    https://en.wikipedia.org/wiki/Test_functions_for_optimization
    """

    def __init__(self):
        self.x = np.zeros(2)

    def get_dofs(self):
        return self.x

    def set_dofs(self, x):
        assert len(x) == 2
        self.x = np.array(x)

    def J(self):
        x = self.x[0]
        y = self.x[1]
        return np.array([1.5 - x + x * y,
                         2.25 - x + x * y * y,
                         2.625 - x + x * y * y * y])



import numpy as np
from jax import grad, vjp
import jax.numpy as jnp
from .jit import jit

from .._core.optimizable import Optimizable


@jit
def curve_length_pure(l):
    """
    This function is used in a Python+Jax implementation of the curve length formula.
    """
    return jnp.mean(l)


class CurveLength(Optimizable):
    r"""
    CurveLength is a class that computes the length of a curve, i.e.

    .. math::
        J = \int_{\text{curve}}~dl.

    """

    def __init__(self, curve):
        self.curve = curve
        self.thisgrad = jit(lambda l: grad(curve_length_pure)(l))
        self.depends_on = ["curve"]

    def J(self):
        """
        This returns the value of the quantity.
        """
        return curve_length_pure(self.curve.incremental_arclength())

    def dJ(self):
        """
        This returns the derivative of the quantity with respect to the curve dofs.
        """
        return self.curve.dincremental_arclength_by_dcoeff_vjp(self.thisgrad(self.curve.incremental_arclength()))


@jit
def Lp_curvature_pure(kappa, gammadash, p, desired_kappa):
    """
    This function is used in a Python+Jax implementation of the curvature penalty term.
    """
    arc_length = jnp.linalg.norm(gammadash, axis=1)
    return (1./p)*jnp.mean(jnp.maximum(kappa-desired_kappa, 0)**p * arc_length)


class LpCurveCurvature(Optimizable):
    r"""
    This class computes a penalty term based on the :math:`L_p` norm
    of the curve's curvature, and penalizes where the local curve curvature exceeds a threshold


    .. math::
        J = \frac{1}{p} \int_{\text{curve}} \text{max}(\kappa - \kappa_0, 0)^p ~dl

    where :math:`\kappa_0` is a threshold curvature.  When ``desired_length`` is
    provided, :math:`\kappa_0 = 2 \pi / \text{desired_length}`, otherwise :math:`\kappa_0=0`.
    """

    def __init__(self, curve, p, desired_length=None):
        self.curve = curve
        self.depends_on = ["curve"]
        if desired_length is None:
            self.desired_kappa = 0
        else:
            radius = desired_length/(2*np.pi)
            self.desired_kappa = 1/radius

        self.J_jax = jit(lambda kappa, gammadash: Lp_curvature_pure(kappa, gammadash, p, self.desired_kappa))
        self.thisgrad0 = jit(lambda kappa, gammadash: grad(self.J_jax, argnums=0)(kappa, gammadash))
        self.thisgrad1 = jit(lambda kappa, gammadash: grad(self.J_jax, argnums=1)(kappa, gammadash))

    def J(self):
        """
        This returns the value of the quantity.
        """
        return self.J_jax(self.curve.kappa(), self.curve.gammadash())

    def dJ(self):
        """
        This returns the derivative of the quantity with respect to the curve dofs.
        """
        grad0 = self.thisgrad0(self.curve.kappa(), self.curve.gammadash())
        grad1 = self.thisgrad1(self.curve.kappa(), self.curve.gammadash())
        return self.curve.dkappa_by_dcoeff_vjp(grad0) + self.curve.dgammadash_by_dcoeff_vjp(grad1)


@jit
def Lp_torsion_pure(torsion, gammadash, p):
    """
    This function is used in a Python+Jax implementation of the formula for the torsion penalty term.
    """
    arc_length = jnp.linalg.norm(gammadash, axis=1)
    return (1./p)*jnp.mean(jnp.abs(torsion)**p * arc_length)


class LpCurveTorsion(Optimizable):
    r"""
    LpCurveTorsion is a class that computes a penalty term based on the :math:`L_p` norm
    of the curve's torsion:


    .. math::
        J = \frac{1}{p} \int_{\text{curve}} \tau^p ~dl.

    """

    def __init__(self, curve, p):
        self.curve = curve
        self.depends_on = ["curve"]

        self.J_jax = jit(lambda torsion, gammadash: Lp_torsion_pure(torsion, gammadash, p))
        self.thisgrad0 = jit(lambda torsion, gammadash: grad(self.J_jax, argnums=0)(torsion, gammadash))
        self.thisgrad1 = jit(lambda torsion, gammadash: grad(self.J_jax, argnums=1)(torsion, gammadash))

    def J(self):
        """
        This returns the value of the quantity.
        """
        return self.J_jax(self.curve.torsion(), self.curve.gammadash())

    def dJ(self):
        """
        This returns the derivative of the quantity with respect to the curve dofs.
        """
        grad0 = self.thisgrad0(self.curve.torsion(), self.curve.gammadash())
        grad1 = self.thisgrad1(self.curve.torsion(), self.curve.gammadash())
        return self.curve.dtorsion_by_dcoeff_vjp(grad0) + self.curve.dgammadash_by_dcoeff_vjp(grad1)


def distance_pure(gamma1, l1, gamma2, l2, minimum_distance):
    """
    This function is used in a Python+Jax implementation of the distance formula.
    """
    dists = jnp.sqrt(jnp.sum((gamma1[:, None, :] - gamma2[None, :, :])**2, axis=2))
    alen = jnp.linalg.norm(l1, axis=1) * jnp.linalg.norm(l2, axis=1)
    return jnp.sum(alen * jnp.maximum(minimum_distance-dists, 0)**2)/(gamma1.shape[0]*gamma2.shape[0])


class MinimumDistance(Optimizable):
    r"""
    MinimumDistance is a class that computes

    .. math::
        J = \sum_{i = 1}^{\text{num_coils}} \sum_{j = 1}^{i-1} d_{i,j}

    where 

    .. math::
        d_{i,j} = \int_{\text{curve}_i} \int_{\text{curve}_j} \max(0, d_{\min} - \| \mathbf{r}_i - \mathbf{r}_j \|_2)^2 ~dl_j ~dl_i\\

    and :math:`\mathbf{r}_i`, :math:`\mathbf{r}_j` are points on coils :math:`i` and :math:`j`, respectively.
    :math:`d_\min` is a desired threshold minimum intercoil distance.  This penalty term is zero when the points on coil :math:`i` and 
    coil :math:`j` lie more than :math:`d_\min` away from one another, for :math:`i, j \in \{1, \cdots, \text{num_coils}\}`

    """

    def __init__(self, curves, minimum_distance):
        self.curves = curves
        self.depends_on = ["curves"]
        self.minimum_distance = minimum_distance

        self.J_jax = jit(lambda gamma1, l1, gamma2, l2: distance_pure(gamma1, l1, gamma2, l2, minimum_distance))
        self.thisgrad0 = jit(lambda gamma1, l1, gamma2, l2: grad(self.J_jax, argnums=0)(gamma1, l1, gamma2, l2))
        self.thisgrad1 = jit(lambda gamma1, l1, gamma2, l2: grad(self.J_jax, argnums=1)(gamma1, l1, gamma2, l2))
        self.thisgrad2 = jit(lambda gamma1, l1, gamma2, l2: grad(self.J_jax, argnums=2)(gamma1, l1, gamma2, l2))
        self.thisgrad3 = jit(lambda gamma1, l1, gamma2, l2: grad(self.J_jax, argnums=3)(gamma1, l1, gamma2, l2))

    def J(self):
        """
        This returns the value of the quantity.
        """
        res = 0
        for i in range(len(self.curves)):
            gamma1 = self.curves[i].gamma()
            l1 = self.curves[i].gammadash()
            for j in range(i):
                gamma2 = self.curves[j].gamma()
                l2 = self.curves[j].gammadash()
                res += self.J_jax(gamma1, l1, gamma2, l2)
        return res

    def dJ(self):
        """
        This returns the derivative of the quantity with respect to the curve dofs.
        """
        dgamma_by_dcoeff_vjp_vecs = [None for c in self.curves]
        dgammadash_by_dcoeff_vjp_vecs = [None for c in self.curves]
        for i in range(len(self.curves)):
            gamma1 = self.curves[i].gamma()
            l1 = self.curves[i].gammadash()
            for j in range(i):
                gamma2 = self.curves[j].gamma()
                l2 = self.curves[j].gammadash()

                temp = self.thisgrad0(gamma1, l1, gamma2, l2)
                if dgamma_by_dcoeff_vjp_vecs[i] is None:
                    dgamma_by_dcoeff_vjp_vecs[i] = temp
                else:
                    dgamma_by_dcoeff_vjp_vecs[i] += temp

                temp = self.thisgrad1(gamma1, l1, gamma2, l2)
                if dgammadash_by_dcoeff_vjp_vecs[i] is None:
                    dgammadash_by_dcoeff_vjp_vecs[i] = temp
                else:
                    dgammadash_by_dcoeff_vjp_vecs[i] += temp

                temp = self.thisgrad2(gamma1, l1, gamma2, l2)
                if dgamma_by_dcoeff_vjp_vecs[j] is None:
                    dgamma_by_dcoeff_vjp_vecs[j] = temp
                else:
                    dgamma_by_dcoeff_vjp_vecs[j] += temp
                temp = self.thisgrad3(gamma1, l1, gamma2, l2)
                if dgammadash_by_dcoeff_vjp_vecs[j] is None:
                    dgammadash_by_dcoeff_vjp_vecs[j] = temp
                else:
                    dgammadash_by_dcoeff_vjp_vecs[j] += temp

        res = [self.curves[i].dgamma_by_dcoeff_vjp(dgamma_by_dcoeff_vjp_vecs[i]) + self.curves[i].dgammadash_by_dcoeff_vjp(dgammadash_by_dcoeff_vjp_vecs[i]) for i in range(len(self.curves))]
        return res



from math import pi

import numpy as np
from jax.ops import index, index_add
import jax.numpy as jnp

from .curve import JaxCurve
from .curve import Curve
import simsoptpp as sopp


class CurveXYZFourier(sopp.CurveXYZFourier, Curve):

    r"""
       CurveXYZFourier is a curve that is represented in Cartesian
       coordinates using the following Fourier series:

        .. math::
           x(\phi) &= \sum_{m=0}^{\text{order}} x_{c,m}\cos(m\phi) + \sum_{m=1}^{\text{order}} x_{s,m}\sin(m\phi) \\
           y(\phi) &= \sum_{m=0}^{\text{order}} y_{c,m}\cos(m\phi) + \sum_{m=1}^{\text{order}} y_{s,m}\sin(m\phi) \\
           z(\phi) &= \sum_{m=0}^{\text{order}} z_{c,m}\cos(m\phi) + \sum_{m=1}^{\text{order}} z_{s,m}\sin(m\phi)

       The dofs are stored in the order

        .. math::
           [x_{c,0}, x_{c,1}, \cdots, x_{c,\text{order}}, x_{s,1}, \cdots, x_{s,\text{order}}, y_{c,0}, \cdots]
    """

    def __init__(self, quadpoints, order):
        if isinstance(quadpoints, int):
            quadpoints = list(np.linspace(0, 1, quadpoints, endpoint=False))
        elif isinstance(quadpoints, np.ndarray):
            quadpoints = list(quadpoints)
        sopp.CurveXYZFourier.__init__(self, quadpoints, order)
        Curve.__init__(self)

    def get_dofs(self):
        """
        This function returns the dofs associated to this object.
        """
        return np.asarray(sopp.CurveXYZFourier.get_dofs(self))

    def set_dofs(self, dofs):
        """
        This function sets the dofs associated to this object.
        """
        sopp.CurveXYZFourier.set_dofs(self, dofs)
        for d in self.dependencies:
            d.invalidate_cache()

    @staticmethod
    def load_curves_from_file(filename, order=None, ppp=20, delimiter=','):
        """
        This function loads a file containing Fourier coefficients for several coils.
        The file is expected to have :mod:`6*num_coils` many columns, and :mod:`order+1` many rows.
        The columns are in the following order,

            sin_x_coil1, cos_x_coil1, sin_y_coil1, cos_y_coil1, sin_z_coil1, cos_z_coil1, sin_x_coil2, cos_x_coil2, sin_y_coil2, cos_y_coil2, sin_z_coil2, cos_z_coil2,  ...

        """
        coil_data = np.loadtxt(filename, delimiter=delimiter)

        assert coil_data.shape[1] % 6 == 0
        assert order <= coil_data.shape[0]-1

        num_coils = coil_data.shape[1]//6
        coils = [CurveXYZFourier(order*ppp, order) for i in range(num_coils)]
        for ic in range(num_coils):
            dofs = coils[ic].dofs
            dofs[0][0] = coil_data[0, 6*ic + 1]
            dofs[1][0] = coil_data[0, 6*ic + 3]
            dofs[2][0] = coil_data[0, 6*ic + 5]
            for io in range(0, min(order, coil_data.shape[0]-1)):
                dofs[0][2*io+1] = coil_data[io+1, 6*ic + 0]
                dofs[0][2*io+2] = coil_data[io+1, 6*ic + 1]
                dofs[1][2*io+1] = coil_data[io+1, 6*ic + 2]
                dofs[1][2*io+2] = coil_data[io+1, 6*ic + 3]
                dofs[2][2*io+1] = coil_data[io+1, 6*ic + 4]
                dofs[2][2*io+2] = coil_data[io+1, 6*ic + 5]
            coils[ic].set_dofs(np.concatenate(dofs))
        return coils


def jaxfouriercurve_pure(dofs, quadpoints, order):
    k = len(dofs)//3
    coeffs = [dofs[:k], dofs[k:(2*k)], dofs[(2*k):]]
    points = quadpoints
    gamma = np.zeros((len(points), 3))
    for i in range(3):
        gamma = index_add(gamma, index[:, i], coeffs[i][0])
        for j in range(1, order+1):
            gamma = index_add(gamma, index[:, i], coeffs[i][2*j-1] * jnp.sin(2*pi*j*points))
            gamma = index_add(gamma, index[:, i], coeffs[i][2*j] * jnp.cos(2*pi*j*points))
    return gamma


class JaxCurveXYZFourier(JaxCurve):

    """
    A Python+Jax implementation of the CurveXYZFourier class.  There is
    actually no reason why one should use this over the C++ implementation in
    :mod:`simsoptpp`, but the point of this class is to illustrate how jax can be used
    to define a geometric object class and calculate all the derivatives (both
    with respect to dofs and with respect to the angle :math:`\phi`) automatically.
    """

    def __init__(self, quadpoints, order):
        if isinstance(quadpoints, int):
            quadpoints = np.linspace(0, 1, quadpoints, endpoint=False)
        pure = lambda dofs, points: jaxfouriercurve_pure(dofs, points, order)
        self.order = order
        self.coefficients = [np.zeros((2*order+1,)), np.zeros((2*order+1,)), np.zeros((2*order+1,))]
        super().__init__(quadpoints, pure)

    def num_dofs(self):
        """
        This function returns the number of dofs associated to this object.
        """
        return 3*(2*self.order+1)

    def get_dofs(self):
        """
        This function returns the dofs associated to this object.
        """
        return np.concatenate(self.coefficients)

    def set_dofs_impl(self, dofs):
        """
        This function sets the dofs associated to this object.
        """
        counter = 0
        for i in range(3):
            self.coefficients[i][0] = dofs[counter]
            counter += 1
            for j in range(1, self.order+1):
                self.coefficients[i][2*j-1] = dofs[counter]
                counter += 1
                self.coefficients[i][2*j] = dofs[counter]
                counter += 1



import numpy as np
import simsoptpp as sopp


class MagneticField(sopp.MagneticField):

    '''
    Generic class that represents any magnetic field from which each magnetic
    field class inherits. The usage of ``MagneticField`` is as follows:

    .. code-block::

        bfield = BiotSavart(coils, currents) # An instance of a MagneticField
        points = ... # points is a (n, 3) numpy array
        bfield.set_points(points)
        B = bfield.B() # returns the Magnetic field at `points`
        dA = bfield.dA_by_dX() # returns the gradient of the potential of the field at `points`

    ``MagneticField`` has a cache to avoid repeated calculations.
    To clear this cache, call the `clear_cached_properties()` function.
    The cache is automatically cleard when ``set_points`` is called.

    '''

    def __init__(self):
        sopp.MagneticField.__init__(self)

    def clear_cached_properties(self):
        """Clear the cache."""
        sopp.MagneticField.invalidate_cache(self)

    def __add__(self, other):
        """Add two magnetic fields."""
        return MagneticFieldSum([self, other])

    def __mul__(self, other):
        """Multiply a field with a scalar."""
        return MagneticFieldMultiply(other, self)

    def __rmul__(self, other):
        """Multiply a field with a scalar."""
        return MagneticFieldMultiply(other, self)

    def to_vtk(self, filename, nr=10, nphi=10, nz=10, rmin=1.0, rmax=2.0, zmin=-0.5, zmax=0.5):
        """Export the field evaluated on a regular grid for visualisation with e.g. Paraview."""
        from pyevtk.hl import gridToVTK
        rs = np.linspace(rmin, rmax, nr, endpoint=True)
        phis = np.linspace(0, 2*np.pi, nphi, endpoint=True)
        zs = np.linspace(zmin, zmax, nz, endpoint=True)

        R, Phi, Z = np.meshgrid(rs, phis, zs)
        X = R * np.cos(Phi)
        Y = R * np.sin(Phi)
        Z = Z

        RPhiZ = np.zeros((R.size, 3))
        RPhiZ[:, 0] = R.flatten()
        RPhiZ[:, 1] = Phi.flatten()
        RPhiZ[:, 2] = Z.flatten()

        self.set_points_cyl(RPhiZ)
        vals = self.B().reshape((R.shape[0], R.shape[1], R.shape[2], 3))
        contig = np.ascontiguousarray
        gridToVTK(filename, X, Y, Z, pointData={"B": (contig(vals[..., 0]), contig(vals[..., 1]), contig(vals[..., 2]))})


class MagneticFieldMultiply(MagneticField):
    """
    Class used to multiply a magnetic field by a scalar.  It takes as input a
    MagneticField class and a scalar and multiplies B, A and their derivatives
    by that value.
    """

    def __init__(self, scalar, Bfield):
        MagneticField.__init__(self)
        self.scalar = scalar
        self.Bfield = Bfield

    def _set_points_cb(self):
        self.Bfield.set_points_cart(self.get_points_cart_ref())

    def _B_impl(self, B):
        B[:] = self.scalar*self.Bfield.B()

    def _dB_by_dX_impl(self, dB):
        dB[:] = self.scalar*self.Bfield.dB_by_dX()

    def _d2B_by_dXdX_impl(self, ddB):
        ddB[:] = self.scalar*self.Bfield.d2B_by_dXdX()

    def _A_impl(self, A):
        A[:] = self.scalar*self.Bfield.A()

    def _dA_by_dX_impl(self, dA):
        dA[:] = self.scalar*self.Bfield.dA_by_dX()

    def _d2A_by_dXdX_impl(self, ddA):
        ddA[:] = self.scalar*self.Bfield.d2A_by_dXdX()


class MagneticFieldSum(MagneticField):
    """
    Class used to sum two or more magnetic field together.  It can either be
    called directly with a list of magnetic fields given as input and outputing
    another magnetic field with B, A and its derivatives added together or it
    can be called by summing magnetic fields classes as Bfield1 + Bfield1
    """

    def __init__(self, Bfields):
        MagneticField.__init__(self)
        self.Bfields = Bfields

    def _set_points_cb(self):
        for bf in self.Bfields:
            bf.set_points_cart(self.get_points_cart_ref())

    def _B_impl(self, B):
        B[:] = np.sum([bf.B() for bf in self.Bfields], axis=0)

    def _dB_by_dX_impl(self, dB):
        dB[:] = np.sum([bf.dB_by_dX() for bf in self.Bfields], axis=0)

    def _d2B_by_dXdX_impl(self, ddB):
        ddB[:] = np.sum([bf.d2B_by_dXdX() for bf in self.Bfields], axis=0)

    def _A_impl(self, A):
        A[:] = np.sum([bf.A() for bf in self.Bfields], axis=0)

    def _dA_by_dX_impl(self, dA):
        dA[:] = np.sum([bf.dA_by_dX() for bf in self.Bfields], axis=0)

    def _d2A_by_dXdX_impl(self, ddA):
        ddA[:] = np.sum([bf.d2A_by_dXdX() for bf in self.Bfields], axis=0)



# coding: utf-8
# Copyright (c) HiddenSymmetries Development Team.
# Distributed under the terms of the LGPL License

"""
This module contains small utility functions and classes needed by *_core*
subpackage.
"""

import itertools
from numbers import Integral, Number
from dataclasses import dataclass
from abc import ABCMeta

import numpy as np

from ..util.types import RealArray


def isbool(val):
    """
    Test whether val is any boolean type, either the native python
    ``bool`` or numpy's ``bool_``.
    """
    return isinstance(val, (bool, np.bool_))


def isnumber(val):
    """
    Test whether val is any kind of number, including both native
    python types or numpy types.
    """
    return isinstance(val, Number)


class Struct:
    """
    This class is just a dummy mutable object to which we can add attributes.
    """


def unique(inlist):
    """
    Given a list or tuple, return a list in which all duplicate
    entries have been removed. Unlike a python set, the order of
    entries in the original list will be preserved.  There is surely
    a faster algorithm than the one used here, but this function will
    not be used in performance-critical code.
    """

    outlist = []
    seen = set()
    for j in inlist:
        if j not in seen:
            outlist.append(j)
            seen.add(j)
    return outlist


@dataclass(frozen=True)
class ImmutableId:
    """
    Immutable class with a single attribute id to represent instance ids. Used
    in conjuction with InstanceCounterMeta metaclass to generate immutable
    instance ids starting with 1 for each of the different classes sublcassing
    InstanceCounterMeta
    """
    id: Integral


class InstanceCounterMeta(type):
    """
    Metaclass to make instance counter not share count with descendants

    Ref: https://stackoverflow.com/questions/8628123/counting-instances-of-a-class
    Credits: https://stackoverflow.com/users/3246302/ratiotile
    """
    def __init__(cls, name, bases, attrs):
        super().__init__(name, bases, attrs)
        cls._ids = itertools.count(1)


class RegisterMeta(type):
    """
    RegisterMeta class can be used to register functions with easy to identify
    names.

    Note:
        The class is not used anymore, but kept to explore the idea 3 explained
        below


    The functionality of RegisterMeta is explained with the Spec class
    defined in simsopt.mhd.spec module. Spec class, which is a subclass
    of Optimizable, implements two functions volume and iota, which are
    used by child Optimizables nodes.

    One could register the two functions of Spec class in a couple of ways.

    1. .. code-block:: python

            Spec.return_fn_map = {'volume': Spec.volume, 'iota': Spec.iota}

    2. .. code-block:: python

            Spec.volume = Spec.register_return_fn("volume")(Spec.volume)
            Spec.iota = Spec.register_return_fn("iota")(Spec.iota)

    3. TO BE IMPLEMENTED

       .. code-block:: python

            class Spec
                ...

                @register_return_fn("volume")
                def volume(self, ...):
                    ...

                @register_return_fn("iota")
                def iota(self, ...):
                    ...
    """
    def __init__(cls, name, bases, attrs):
        super().__init__(name, bases, attrs)
        cls.return_fn_map = {}

        def _register_return_fn(name):
            def inner_register(f):
                cls.return_fn_map[name] = f
                return f
            return inner_register

        cls.register_return_fn = _register_return_fn


#class OptimizableMeta(InstanceCounterMeta, RegisterMeta, ABCMeta):
class OptimizableMeta(InstanceCounterMeta, ABCMeta):
    """
    Meta class for Optimizable class
    """
    pass


class ObjectiveFailure(Exception):
    """
    Defines a custom exception used to indicate failure when
    evaluating the objective function. For example, if Vmec or Spec
    fail to converge, this exception will be thrown. The simsopt
    solvers will catch this specific exception (not others) and set
    the objective function to a large number.
    """
    pass


def finite_difference_steps(x: RealArray,
                            abs_step: float = 1.0e-7,
                            rel_step: float = 0.0
                            ) -> RealArray:
    """
    Determine an array of step sizes for calculating finite-difference
    derivatives, using absolute or relative step sizes, or a mixture
    thereof.

    For each element ``x_j`` of the state vector ``x``, the step size
    ``s_j`` is determined from

    ``s_j = max(abs(x_j) * rel_step, abs_step)``

    So, if you want to use the same absolute step size for all
    elements of ``x``, set ``rel_step = 0`` and set ``abs_step``
    positive. Or, if you want to use the same relative step size for
    all elements of ``x``, set ``abs_step = 0`` and ``rel_step``
    positive. If both ``abs_step`` and ``rel_step`` are positive, then
    ``abs_step`` effectively gives the lower bound on the step size.

    It is dangerous to set ``abs_step`` to exactly 0, since then if
    any elements of ``x`` are 0, the step size will be 0. In this
    situation, ``ValueError`` will be raised. It is preferable for
    ``abs_step`` to be a small positive value.

    For one-sided finite differences, the values of ``x_j`` used will
    be ``x_j + s_j``. For centered finite differences, the values of
    ``x_j`` used will be ``x_j + s_j`` and ``x_j - s_j``.

    This routine is used by :func:`simsopt._core.dofs.fd_jac()` and
    :func:`simsopt.solve.mpi.fd_jac_mpi()`.

    Args:
        x: The state vector at which you wish to compute the gradient
          or Jacobian. Must be a 1D array.
        abs_step: The absolute step size.
        rel_step: The relative step size.

    Returns:
        A 1D numpy array of the same size as ``x``, with each element
        being the step size used for each corresponding element of
        ``x``.
    """
    if abs_step < 0:
        raise ValueError('abs_step must be >= 0')
    if rel_step < 0:
        raise ValueError('rel_step must be >= 0')

    steps = np.max((np.abs(x) * rel_step, np.full(len(x), abs_step)),
                   axis=0)

    # If abs_step == 0 and any elements of x are 0, we could end up
    # with a step of size 0:
    if np.any(steps == 0.0):
        raise ValueError('Finite difference step size cannot be 0. ' \
                         'Increase abs_step.')

    return steps



############################################################################
# Copyright (c) Wolf Vollprecht, Johan Mabille and Sylvain Corlay          #
# Copyright (c) QuantStack                                                 #
#                                                                          #
# Distributed under the terms of the BSD 3-Clause License.                 #
#                                                                          #
# The full license is in the file LICENSE, distributed with this software. #
############################################################################

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
import sys
import os
import setuptools

__version__ = '0.0.1'


class get_pybind_include(object):
    """Helper class to determine the pybind11 include path

    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11
        return pybind11.get_include(self.user)

class get_numpy_include(object):
    """Helper class to determine the numpy include path

    The purpose of this class is to postpone importing numpy
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __str__(self):
        import numpy
        return numpy.get_include()

ext_modules = [
    Extension(
        'xtensor_python_test',
        ['main.cpp'],
        include_dirs=[
            # Path to pybind11 headers
            '../include/',
            get_pybind_include(),
            get_pybind_include(user=True),
            # Path to numpy headers
            get_numpy_include(),
            os.path.join(sys.prefix, 'include'),
            os.path.join(sys.prefix, 'Library', 'include')
        ],
        language='c++'
    ),
]


def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile
    with tempfile.NamedTemporaryFile('w', suffix='.cpp') as f:
        f.write('int main (int argc, char **argv) { return 0; }')
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def cpp_flag(compiler):
    """Return the -std=c++14 compiler flag  and errors when the flag is
    no available.
    """
    if has_flag(compiler, '-std=c++14'):
        return '-std=c++14'
    else:
        raise RuntimeError('C++14 support is required by xtensor!')


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""
    c_opts = {
        'msvc': ['/EHsc'],
        'unix': [],
    }

    if sys.platform == 'darwin':
        c_opts['unix'] += ['-stdlib=libc++', '-mmacosx-version-min=10.7']

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        if ct == 'unix':
            opts.append('-DVERSION_INFO="%s"' % self.distribution.get_version())
            opts.append(cpp_flag(self.compiler))
            if has_flag(self.compiler, '-fvisibility=hidden'):
                opts.append('-fvisibility=hidden')
        elif ct == 'msvc':
            opts.append('/DVERSION_INFO=\\"%s\\"' % self.distribution.get_version())
        for ext in self.extensions:
            ext.extra_compile_args = opts
        build_ext.build_extensions(self)

setup(
    name='xtensor_python_test',
    version=__version__,
    author='Sylvain Corlay',
    author_email='sylvain.corlay@gmail.com',
    url='https://github.com/pybind/python_example',
    description='An example project using xtensor-python',
    long_description='',
    ext_modules=ext_modules,
    install_requires=['pybind11>=2.0.1'],
    cmdclass={'build_ext': BuildExt},
    zip_safe=False,
)



# coding: utf-8
# Copyright (c) HiddenSymmetries Development Team.
# Distributed under the terms of the LGPL License

"""
This module contains the :class:`~simsopt.util.mpi.MpiPartition` class.

This module should be completely self-contained, depending only on
mpi4py and numpy, not on any other simsopt components.
"""
__all__ = ['log', 'MpiPartition']

import logging
from typing import Union
import numpy as np

from monty.dev import requires
try:
    from mpi4py import MPI
except ImportError:
    MPI = None

STOP = 0


def log(level: int = logging.INFO):
    """
    Turn on logging. If MPI is available, the processor number will be
    added to all logging entries.

    Args:
        level: Typically ``logging.INFO`` for regular output, or 
          ``logging.DEBUG`` for more extensive output.
    """
    format = "%(levelname)s:%(name)s:%(lineno)d %(message)s"
    if MPI is not None:
        format = "[{}] ".format(MPI.COMM_WORLD.Get_rank()) + format

    logging.basicConfig(level=level, format=format)


logger = logging.getLogger(__name__)


@requires(MPI is not None, "mpi4py is not installed")
class MpiPartition:
    """
    This module contains functions related to dividing up the set of
    MPI processes into groups, each of which can work together. For
    more information, see :ref:`mpi`.

    Args:
        ngroups: The number of worker groups desired. If ``None``, a worker
          group will be created for each MPI process. If a value is supplied
          that is larger than the number of processes, the number will be
          lowered to the number of processes.
        comm_world: The MPI communicator containing all processes to split into
          worker groups.

    Attributes:
        ngroups (int): The number of worker groups.
        group (int): Index giving the worker group that this MPI process belongs to.
        comm_world (mpi4py.MPI.Intracomm): The MPI communicator that includes all processes together.
        comm_groups (mpi4py.MPI.Intracomm): The MPI communicator representing the worker groups.
        comm_leaders (mpi4py.MPI.Intracomm): The MPI communicator that includes only leaders of worker groups.
        rank_world (int): The MPI rank in the ``comm_world`` communicator.
        rank_groups (int): The MPI rank in the ``comm_groups`` communicator.
        rank_leaders (int): The MPI rank in the ``comm_leaders`` communicator, or -1 for processes that are not group leaders.
        nprocs_world (int): The number of MPI processes in the ``comm_world`` communicator.
        nprocs_groups (int): The number of MPI processes in this process's worker group.
        nprocs_leaders (int): The number of group leaders, if this process is a group leader, otherwise -1.
        proc0_world (bool): Whether this MPI process has rank 0 in ``comm_world``.
        proc0_groups (bool): Whether this MPI process has rank 0 in ``comm_groups``, i.e. whether this process is a group leader.
    .
    """

    def __init__(self,
                 # ngroups: Union[None, int] = None,
                 # comm_world: Union[MPI.Intracomm, None] = MPI.COMM_WORLD):
                 ngroups=None,
                 comm_world=None):
        # if MPI is None:
        #     raise RuntimeError("MpiPartition class requires the mpi4py package.")

        self.is_apart = False
        self.comm_world = comm_world if comm_world is not None else MPI.COMM_WORLD

        self.rank_world = self.comm_world.Get_rank()
        self.nprocs_world = self.comm_world.Get_size()
        self.proc0_world = (self.rank_world == 0)

        if ngroups is None:
            ngroups = self.nprocs_world
        # Force ngroups to be in the range [1, nprocs_world]
        if ngroups < 1:
            ngroups = 1
            logger.info('Raising ngroups to 1')
        if ngroups > self.nprocs_world:
            ngroups = self.nprocs_world
            logger.info('Lowering ngroups to {}'.format(ngroups))
        self.ngroups = ngroups

        self.group = int(
            np.floor((self.rank_world * ngroups) / self.nprocs_world))

        # Set up the "groups" communicator:
        self.comm_groups = self.comm_world.Split(color=self.group,
                                                 key=self.rank_world)
        self.rank_groups = self.comm_groups.Get_rank()
        self.nprocs_groups = self.comm_groups.Get_size()
        self.proc0_groups = (self.rank_groups == 0)

        # Set up the "leaders" communicator:
        if self.proc0_groups:
            color = 0
        else:
            color = MPI.UNDEFINED
        self.comm_leaders = self.comm_world.Split(color=color,
                                                  key=self.rank_world)
        if self.proc0_groups:
            self.rank_leaders = self.comm_leaders.Get_rank()
            self.nprocs_leaders = self.comm_leaders.Get_size()
        else:
            # We are not allowed to query the rank from procs that are
            # not members of comm_leaders.
            self.rank_leaders = -1
            self.nprocs_leaders = -1

    def write(self):
        """ Print info about the MPI configuration """
        columns = ["rank_world", "nprocs_world", "group", "ngroups",
                   "rank_groups", "nprocs_groups", "rank_leaders",
                   "nprocs_leaders"]
        data = [self.rank_world, self.nprocs_world, self.group, self.ngroups,
                self.rank_groups, self.nprocs_groups, self.rank_leaders,
                self.nprocs_leaders]

        # Each processor sends their data to proc0_world, and
        # proc0_world writes the result to the file in order.
        if self.proc0_world:
            # Print header row
            width = max(len(s) for s in columns) + 1
            print(",".join(s.rjust(width) for s in columns))
            print(",".join(str(s).rjust(width) for s in data))
            for tag in range(1, self.nprocs_world):
                data = self.comm_world.recv(tag=tag)
                print(",".join(str(s).rjust(width) for s in data))
        else:
            tag = self.rank_world
            self.comm_world.send(data, 0, tag)

    def mobilize_leaders(self, action_const):
        """
        This function is called by ``proc0_world`` to tell the other
        group leaders that it is time to begin some action,
        e.g. starting to calculat a finite difference Jacobian.
        """
        logger.debug('mobilize_leaders, action_const={}'.format(action_const))
        if not self.proc0_world:
            raise RuntimeError(
                'Only proc0_world should call mobilize_leaders()')

        self.comm_leaders.bcast(action_const, root=0)

    def mobilize_workers(self, action_const):
        logger.debug('mobilize_workers, action_const={}'.format(action_const))
        if not self.proc0_groups:
            raise RuntimeError(
                'Only group leaders should call mobilize_workers()')

        self.comm_groups.bcast(action_const, root=0)

    def stop_leaders(self):
        logger.debug('stop_leaders')
        if not self.proc0_world:
            raise RuntimeError('Only proc0_world should call stop_leaders()')

        data = STOP
        self.comm_leaders.bcast(data, root=0)

    def stop_workers(self):
        logger.debug('stop_workers')
        if not self.proc0_groups:
            raise RuntimeError('Only proc0_groups should call stop_workers()')

        data = STOP
        self.comm_groups.bcast(data, root=0)

    def leaders_loop(self, action):
        """
        actions should be a dict where the keys are possible integer
        constants, and the values are callable functions that are
        called when the corresponding key is sent from
        mobilize_leaders.
        """
        if self.proc0_world:
            logger.debug('proc0_world bypassing leaders_loop')
            return

        logger.debug('entering leaders_loop')

        while True:
            # Wait for proc 0 to send us something:
            data = None
            data = self.comm_leaders.bcast(data, root=0)
            logger.debug('leaders_loop received {}'.format(data))
            if data == STOP:
                # Tell workers to stop
                break

            # Call the requested function:
            action(self, data)

        logger.debug('leaders_loop end')

    def worker_loop(self, action):
        """
        actions should be a dict where the keys are possible integer
        constants, and the values are callable functions that are
        called when the corresponding key is sent from
        mobilize_workers.
        """
        if self.proc0_groups:
            logger.debug('bypassing worker_loop since proc0_groups')
            return

        logger.debug('entering worker_loop')

        while True:
            # Wait for the group leader to send us something:
            data = None
            data = self.comm_groups.bcast(data, root=0)
            logger.debug('worker_loop worker received {}'.format(data))
            if data == STOP:
                break

            # Call the requested function:
            action(self, data)

        logger.debug('worker_loop end')

    def apart(self, leaders_action, workers_action):
        """
        Send workers and group leaders off to their respective loops to
        wait for instructions from their group leader or
        proc0_world, respectively.
        """
        self.is_apart = True
        if self.proc0_world:
            pass
        elif self.proc0_groups:
            self.leaders_loop(leaders_action)
        else:
            self.worker_loop(workers_action)

    def together(self):
        """
        Bring workers and group leaders back from their respective loops.
        """
        if self.proc0_world:
            self.stop_leaders()  # Proc0_world stops the leaders.

        if self.proc0_groups:
            self.stop_workers()  # All group leaders stop their workers.

        self.is_apart = False



from benchmark_xtensor_python import sum_tensor
import numpy as np

u = np.ones(1000000, dtype=float)
#print(sum_tensor(u))
from timeit import timeit
print (timeit ('sum_tensor(u)', setup='from __main__ import u, sum_tensor', number=1000))



from scipy.optimize import minimize, least_squares, NonlinearConstraint
import numpy as np
from simsopt.geo.surfaceobjectives import QfmResidual


class QfmSurface():
    r"""
    QfmSurface is used to compute a quadratic-flux minimizing surface, defined
    as the minimum of the objective function,

    .. math::
        f(S) = \frac{\int_{S} d^2 x \, \left(\textbf{B} \cdot \hat{\textbf{n}}\right)^2}{\int_{S} d^2 x \, B^2}

    subject to a constraint on the surface label, such as the area, volume,
    or toroidal flux. This objective, :math:`f`, is computed in :class:`QfmResidual`
    (see :mod:`surfaceobjectives.py`). If magnetic surfaces exists, :math:`f`
    will be zero. If not, QFM surfaces approximate flux surfaces.

    The label constraint can be enforced with a penalty formulation, defined
    in :func:`qfm_penalty_constraints()`, and whose minimium is computed with
    LBFGS-B by :func:`minimize_qfm_penalty_constraints_LBFGS()'.

    Alternatively, the label constraint can be enforced with a constrained
    optimization algorithm. This constrained optimization problem is
    solved with the SLSQP algorithm by :func:`minimize_qfm_exact_constraints_SLSQP()`.
    """

    def __init__(self, biotsavart, surface, label, targetlabel):
        self.bs = biotsavart
        self.surface = surface
        self.label = label
        self.targetlabel = targetlabel

    def qfm_label_constraint(self, x, derivatives=0):
        r"""
        Returns the residual

        .. math::
            0.5 (\texttt{label} - \texttt{labeltarget})^2

        and optionally the gradient wrt surface params.
        """
        assert derivatives in [0, 1]

        s = self.surface
        sdofs = x
        nsurfdofs = x.size
        bs = self.bs

        s.set_dofs(sdofs)
        l = self.label.J()
        rl = l-self.targetlabel
        val = 0.5 * rl ** 2

        if derivatives:
            dl = self.label.dJ_by_dsurfacecoefficients()
            dval = rl*dl
            return val, dval
        else:
            return val

    def qfm_objective(self, x, derivatives=0):
        r"""
        Returns the residual

        .. math::
            f(S)

        and optionally the gradient wrt surface params where :math:`f(S)`` is
        the QFM residual defined in surfaceobjectives.py.
        """
        assert derivatives in [0, 1]

        s = self.surface
        sdofs = x
        nsurfdofs = x.size
        bs = self.bs

        s.set_dofs(sdofs)
        qfm = QfmResidual(s, bs)

        r = qfm.J()

        if derivatives == 0:
            return r
        else:
            dr = qfm.dJ_by_dsurfacecoefficients()
            return r, dr

    def qfm_penalty_constraints(self, x, derivatives=0, constraint_weight=1.):
        r"""
        Returns the residual

        .. math::
            f(S) + 0.5 \texttt{constraint_weight} (\texttt{label} - \texttt{labeltarget})^2

        and optionally the gradient and Hessian wrt surface params where
        :math:`f(S)` is the QFM residual defined in :mod:`surfaceobjectives.py`.
        """

        assert derivatives in [0, 1, 2]
        s = self.surface
        sdofs = x
        nsurfdofs = x.size
        bs = self.bs

        s.set_dofs(sdofs)

        qfm = QfmResidual(s, bs)

        r = qfm.J()

        l = self.label.J()
        rl = l-self.targetlabel

        val = r + 0.5 * constraint_weight * rl**2
        if derivatives == 0:
            return val

        dr = qfm.dJ_by_dsurfacecoefficients()

        dl = self.label.dJ_by_dsurfacecoefficients()

        dval = dr + constraint_weight*rl*dl
        if derivatives == 1:
            return val, dval

        d2r = qfm.d2J_by_dsurfacecoefficientsdsurfacecoefficients()
        d2l = self.label.d2J_by_dsurfacecoefficientsdsurfacecoefficients()
        d2val = d2r  \
            + constraint_weight*rl*d2l + constraint_weight*dl[:, None]*dl[None, :]

        return val, dval, d2val

    def minimize_qfm_penalty_constraints_LBFGS(self, tol=1e-3, maxiter=1000,
                                               constraint_weight=1.):
        r"""
        This function tries to find the surface that approximately solves

        .. math::
            \min_S f(S) + 0.5 \texttt{constraint_weight} (\texttt{label} - \texttt{labeltarget})^2

        where :math:`f(S)` is the QFM residual defined in :mod:`surfaceobjectives.py.`
        This is done using LBFGS.
        """

        s = self.surface
        x = s.get_dofs()
        fun = lambda x: self.qfm_penalty_constraints(
            x, derivatives=1, constraint_weight=constraint_weight)
        res = minimize(
            fun, x, jac=True, method='L-BFGS-B',
            options={'maxiter': maxiter, 'ftol': tol, 'gtol': tol,
                     'maxcor': 200})

        resdict = {
            "fun": res.fun, "gradient": res.jac, "iter": res.nit, "info": res,
            "success": res.success,
        }
        s.set_dofs(res.x)
        resdict['s'] = s

        return resdict

    def minimize_qfm_exact_constraints_SLSQP(self, tol=1e-3, maxiter=1000):
        r"""
        This function tries to find the surface that approximately solves

        .. math::
            \min_{S} f(S)

        subject to

        .. math::
            \texttt{label} = \texttt{labeltarget}

        where :math:`f(S)` is the QFM residual. This is done using SLSQP.
        """
        s = self.surface
        x = s.get_dofs()

        fun = lambda x: self.qfm_objective(x, derivatives=1)
        con = lambda x: self.qfm_label_constraint(x, derivatives=1)[0]
        dcon = lambda x: self.qfm_label_constraint(x, derivatives=1)[1]

        nlc = NonlinearConstraint(con, 0, 0)
        eq_constraints = [{'type': 'eq', 'fun': con, 'jac': dcon}]
        res = minimize(
            fun, x, jac=True, method='SLSQP', constraints=eq_constraints,
            options={'maxiter': maxiter, 'ftol': tol})

        resdict = {
            "fun": res.fun, "gradient": res.jac, "iter": res.nit, "info": res,
            "success": res.success,
        }
        s.set_dofs(res.x)
        resdict['s'] = s

        return resdict

    def minimize_qfm(self, tol=1e-3, maxiter=1000, method='SLSQP',
                     constraint_weight=1.):
        r"""
        This function tries to find the surface that approximately solves

        .. math::
            \min_{S} f(S)

        subject to

        .. math::
            \texttt{label} = \texttt{labeltarget}

        where :math:`f(S)` is the QFM residual. This is done using method,
        either 'SLSQP' or 'LBFGS'. If 'LBFGS' is chosen, a penalized formulation
        is used defined by constraint_weight. If 'SLSQP' is chosen, constraint_weight
        is not used, and the constrained optimization problem is solved.
        """
        if method == 'SLSQP':
            return self.minimize_qfm_exact_constraints_SLSQP(tol=tol,
                                                             maxiter=maxiter)
        elif method == 'LBFGS':
            return self.minimize_qfm_penalty_constraints_LBFGS(tol=tol,
                                                               maxiter=maxiter, constraint_weight=constraint_weight)
        else:
            raise ValueError



# coding: utf-8
# Copyright (c) HiddenSymmetries Development Team.
# Distributed under the terms of the LGPL License

"""
This module contains small utility functions and classes.
"""

from typing import Union, Sequence
from numbers import Integral, Real

from nptyping import NDArray, Float, Int, Bool

Array = Union[Sequence, NDArray]
RealArray = Union[Sequence[Real], NDArray[Float]]
IntArray = Union[Sequence[Integral], NDArray[Int]]
StrArray = Union[Sequence[str], NDArray[str]]
BoolArray = Union[Sequence[bool], NDArray[Bool]]
Key = Union[Integral, str]




# coding: utf-8
# Copyright (c) HiddenSymmetries Development Team.
# Distributed under the terms of the LGPL License

"""
This module provides classes and functions that are useful for
setting up optimizable objects and objective functions.
"""

import logging
import types
import numpy as np

logger = logging.getLogger(__name__)


class Optimizable():
    """
    This base class provides some useful features for optimizable functions.
    """

    def get_dofs(self):
        """
        This base Optimizable object has no degrees of freedom, so return
        an empty array
        """
        return np.array([])

    def set_dofs(self, x):
        """
        This base Optimizable object has no degrees of freedom, so do nothing.
        """
        pass

    def index(self, dof_str):
        """
        Given a string dof_str, returns the index in the dof array whose
        name matches dof_str. If dof_str does not match any of the
        names, ValueError will be raised.
        """
        return self.names.index(dof_str)

    def get(self, dof_str):
        """
        Return a degree of freedom specified by its string name.
        """
        x = self.get_dofs()
        return x[self.index(dof_str)]

    def set(self, dof_str, newval):
        """
        Set a degree of freedom specified by its string name.
        """
        x = self.get_dofs()
        x[self.index(dof_str)] = newval
        self.set_dofs(x)

    def get_fixed(self, dof_str):
        """
        Return the fixed attribute for a given degree of freedon, specified by dof_str.
        """
        return self.fixed[self.index(dof_str)]

    def set_fixed(self, dof_str, fixed_new=True):
        """
        Set the fixed attribute for a given degree of freedom, specified by dof_str.
        """
        self.fixed[self.index(dof_str)] = fixed_new

    def all_fixed(self, fixed_new=True):
        """
        Set the 'fixed' attribute for all degrees of freedom.
        """
        self.fixed = np.full(len(self.get_dofs()), fixed_new)


def function_from_user(target):
    """
    Given a user-supplied "target" to be optimized, extract the
    associated callable function.
    """
    if callable(target):
        return target
    elif hasattr(target, 'J') and callable(target.J):
        return target.J
    else:
        raise TypeError('Unable to find a callable function associated '
                        'with the user-supplied target ' + str(target))


class Target(Optimizable):
    """
    Given an attribute of an object, which typically would be a
    @property, form a callable function that can be used as a target
    for optimization.
    """

    def __init__(self, obj, attr):
        self.obj = obj
        self.attr = attr
        self.depends_on = ["obj"]

        # Attach a dJ function only if obj has one
        def dJ(self0):
            return getattr(self0.obj, 'd' + self0.attr)
        if hasattr(obj, 'd' + attr):
            self.dJ = types.MethodType(dJ, self)

    def J(self):
        return getattr(self.obj, self.attr)

    #def dJ(self):
    #    return getattr(self.obj, 'd' + self.attr)

    def get_dofs(self):
        return np.array([])

    def set_dofs(self, v):
        pass


def make_optimizable(obj):
    """
    Given any object, add attributes like fixed, mins, and maxs. fixed
    = False by default. Also, add the other methods of Optimizable to
    the object.
    """

    # If the object does not have a get_dofs() method, attach one,
    # assuming the object does not directly own any dofs.
    def get_dofs(self):
        return np.array([])

    def set_dofs(self, x):
        pass
    if not hasattr(obj, 'get_dofs'):
        obj.get_dofs = types.MethodType(get_dofs, obj)
    if not hasattr(obj, 'set_dofs'):
        obj.set_dofs = types.MethodType(set_dofs, obj)

    n = len(obj.get_dofs())
    if not hasattr(obj, 'fixed'):
        obj.fixed = np.full(n, False)
    if not hasattr(obj, 'mins'):
        obj.mins = np.full(n, np.NINF)
    if not hasattr(obj, 'maxs'):
        obj.maxs = np.full(n, np.Inf)
    # Add the following methods from the Optimizable class:
    for method in ['index', 'get', 'set', 'get_fixed', 'set_fixed', 'all_fixed']:
        # See https://stackoverflow.com/questions/972/adding-a-method-to-an-existing-object-instance
        setattr(obj, method, types.MethodType(getattr(Optimizable, method), obj))

    return obj



# coding: utf-8
# Copyright (c) HiddenSymmetries Development Team.
# Distributed under the terms of the LGPL License

__all__ = ['initialize_logging']

import logging
import logging.config
from pathlib import Path

from ruamel.yaml import YAML
try:
    from mpi4py import MPI
    # from .mpi_logger import MPILogHandler
except:
    MPI = None


def initialize_logging(filename: str = None,
                       level: str = None,
                       mpi: bool = False) -> None:
    """
    Initializes logging in a simple way for both serial and MPI jobs.
    The MPI logging uses MPILogger package.

    Args:
        filename: Name of file to store the logging info
        level: Logging level. Could be 'INFO', 'DEBUG', 'WARNING', etc.
        mpi: If True MPI logging is used provided mpi4py is installed.
    """
    yaml = YAML(typ='safe')
    config_dict = yaml.load(Path(__file__).parent / 'log_config.yaml')
    if filename:
        config_dict['handlers']['file_handler'].update({'filename': filename})
        config_dict['handlers']['mpi_file_handler'].update({'logfile': filename})
    if level:
        config_dict['handlers']['file_handler'].update({'level': level})
        config_dict['handlers']['mpi_file_handler'].update({'level': level})
    if mpi and MPI is not None:
        config_dict['root']['handlers'].pop(2)  # Remove file hander
    else:
        config_dict['root']['handlers'].pop(3)  # Remove mpi hander
        del config_dict['handlers']['mpi_file_handler']

    logging.config.dictConfig(config_dict)

    if mpi and MPI is None:
        logging.warning("mpi4py not installed. Not able to log MPI info")
